#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
CSV and git-based database to store year-country data in a multi-user
setup
"""


import os 
import git
import tqdm
import time
import copy
import types

import traceback
from collections import defaultdict
from pathlib import Path
import shutil

import pandas as pd
import numpy as np

from . import config
from .data_structures import Datatable, TableSet, read_csv
from . import mapping as mapp
from . import io_tools as io
from . import util
from . import core
from .tools import xarray

# helper funtions
def restore_source(repoName):
    """
    Restores the data respository based on the last working git hash in the inventory file.
    After a successfulr restorations, datatoolbox need to be imported from scratch. For notebooks
    the restart of the kernal might be required.

    Parameters
    ----------
    repoName : str
        Source ID.

    Returns
    -------
    None.

    """
    
    #gitManager = GitRepository_Manager(config)
    #sources   = gitManager.sources
    sources = pd.read_csv(config.SOURCE_FILE, index_col='SOURCE_ID')
    git_hash = sources.git_commit_hash.loc[repoName]
    repoPath = os.path.join(config.PATH_TO_DATASHELF,  'database', repoName)
    git_repo = git.Repo(repoPath)
    git_repo.git.execute(['git', 'reset', '--hard', git_hash])
    
def unique(table, field=None):
    #%%
    unique_dict = dict()
    max_len = 0
    if field is None:
        fields = config.ID_FIELDS
    else:
        fields = [field]
    
    for var in fields:
        unique_dict[var] = table.loc[:,var].unique()
        max_len = max(max_len, len(unique_dict[var]))
    unique_data = pd.DataFrame(index = range(max_len),
                              columns = fields,
                              data = '').astype(str)
    
    for var in fields:
        len_data = len(unique_dict[var])
        unique_data.loc[unique_data.index[:len_data],var] = unique_dict[var]
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(unique_data)
        
        
#%%
    
        
class Database():
    """
    CSV based database that uses git for as distributed version control system.
    Each table is saved locally as a csv file and identified by a  unique ID. 
    The csv files are organized in various sources in individual folders. Each 
    sources comes with its own git repository and can be shared with others.
    """
    def __init__(self):
        """ 
        Initialized the database and creates an empty one in case the directory
        in the config is empty.
        """
        tt = time.time()
        self.path = config.PATH_TO_DATASHELF
        
        if not os.path.exists(os.path.join(self.path, 'inventory.csv')):
            self.create_empty_datashelf(config.MODULE_PATH,
                                        self.path)
        
        self.gitManager = GitRepository_Manager(config)
        self.INVENTORY_PATH = os.path.join(self.path, 'inventory.csv')
        self.inventory = pd.read_csv(self.INVENTORY_PATH, index_col=0, dtype={'source_year': str},low_memory=False)
        self.sources   = self.gitManager.sources
        
        #self.gitManager._validateRepository('main')


        if config.DEBUG:
            print('Database loaded in {:2.4f} seconds'.format(time.time()-tt))
    
        if config.TEST_ENV:
            # if no config is given, the empty sandbox is populated with some test
            # tables
            tablesToCommit, source_meta  = util._create_sandbox_tables('SOURCE_A_2020', 1)
            self.commitTables(tablesToCommit,
                         message = 'added dummy tablesof source A', 
                         sourceMetaDict = source_meta)
            tablesToCommit, source_meta  = util._create_sandbox_tables('SOURCE_B_2020', 2)
            self.commitTables(tablesToCommit,
                         message = 'added dummy tables of source B', 
                         sourceMetaDict = source_meta)
            print('Added test tables to Sandbox datashelf')
            
 
    
    def create_empty_datashelf(self,
                               modulePath, 
                               pathToDataself):
        """
        Method to create the required files for an empty csv-based data base. 
        (Equivalent to the fucntions in admin.py)
        """
        from pathlib import Path
        import os
        import shutil
        path = Path(pathToDataself)
        path.mkdir(parents=True,exist_ok=True)
        
        # add subfolders database
        Path(os.path.join(pathToDataself, 'database')).mkdir(exist_ok=True)
        Path(os.path.join(pathToDataself, 'mappings')).mkdir(exist_ok=True)
        Path(os.path.join(pathToDataself, 'rawdata')).mkdir(exist_ok=True)
        
        #create mappings
        os.makedirs(os.path.join(pathToDataself, 'mappings'),exist_ok=True)
        shutil.copyfile(os.path.join(modulePath, 'data/regions.csv'),
                        os.path.join(pathToDataself, 'mappings/regions.csv'))
        shutil.copyfile(os.path.join(modulePath, 'data/continent.csv'),
                        os.path.join(pathToDataself, 'mappings/continent.csv'))
        shutil.copyfile(os.path.join(modulePath, 'data/country_codes.csv'),
                        os.path.join(pathToDataself, 'mappings/country_codes.csv'))    
        
        # created sources.csv that contains the indivitual information for each source
        sourcesDf = pd.DataFrame(columns = config.SOURCE_META_FIELDS)
        filePath= os.path.join(pathToDataself, 'sources.csv')
        sourcesDf.to_csv(filePath)
        
        # creates inventory.csv that contains all data tables from all sources
        inventoryDf = pd.DataFrame(columns = config.INVENTORY_FIELDS)
        filePath= os.path.join(pathToDataself, 'inventory.csv')
        inventoryDf.to_csv(filePath)
        git.Repo.init(pathToDataself)
    

        
    def _validateRepository(self, repoID='main'):
        """
        Private
        Checks that a sub repository is valid. Valid means that the git repository
        is clean and not outanding commits are there.
        """
        return self.gitManager._validateRepository(repoID)
    
    def info(self):
        """
        Shows the most inmportant information about the status of the database
        """
        
        print('######## Database informations: #############')
        print('Your database is located at: ' + self.path)
        print('Number of tables: {}'.format(len(self.inventory)))  
        print('Number of data sources: {}'.format(len(self.gitManager.sources))) 
        print('Number of commits: {}'.format(self.gitManager['main'].git.rev_list("--all", "--count")))
#        root_directory = Path(config.PATH_TO_DATASHELF)
#        print('Size of datashelf: {:2.2f} MB'.format(sum(f.stat().st_size for f in root_directory.glob('**/*') if f.is_file() )/1e6))
#        root_directory = Path(os.path.join(config.PATH_TO_DATASHELF, 'rawdata'))
#        print('Size of raw_data: {:2.2f} MB'.format(sum(f.stat().st_size for f in root_directory.glob('**/*') if f.is_file() )/1e6))
        print('#############################################')
        
    def sourceInfo(self):
        """ 
        Returns a list of available sources and meta data
        """
        sources = copy.copy(self.gitManager.sources)
        return sources.sort_index()

    def returnInventory(self):
        """
        Returns a copy of the data base inventory. 
        """
        return copy.copy(self.inventory)


    def _reloadInventory(self):
        """
        Private
        Reloades the inventory from the csv file
        """
        self.inventory = pd.read_csv(self.INVENTORY_PATH, index_col=0, dtype={'source_year': str})
        
    def sourceExists(self, source):
        """ 
        Function to check if a source is propperly registered in the data base
        
        Input: SourceID
        """
        return source in self.gitManager.sources.index
    
    def add_to_inventory(self, datatable):
        """
        Method to add a table to the global inventory file. 
        Input: datatable
        """
        self.inventory.loc[datatable.ID] = [datatable.meta.get(x,None) for x in config.INVENTORY_FIELDS]
        #self.gitManager.updatedRepos.add('main') # TODO not needed anymore?
        

    def remove_from_inventory(self, tableID):
        """
        Method to remove a table from the global inventory
        Input: tableID
        """
        self.inventory.drop(tableID, inplace=True)
    
    def findc(self, **kwargs):
        """
        Method to search through the inventory. kwargs can be all inventory entires
        (see config.INVENTORY_FIELDS).
        """
        
        # loop over all keys of kwargs to filter based on all of them
        mask = True
        for key, val in kwargs.items():
            mask &= self.inventory[key].str.contains(val, regex=False, na=False)

        table = (
            self.inventory if mask is True else self.inventory.loc[mask]
        ).copy()
        
        # test to add function to a instance (does not require class)
        table.graph = types.MethodType(util.plot_query_as_graph, table)
        table.short = types.MethodType(util.shorten_find_output, table)
        table.unique = types.MethodType(unique, table)
        table.to_pyam = types.MethodType(util.to_pyam, table)
        return table

    def findp(self, level=None, regex=False, **filters):
        """ 
        Future defaulf find method that allows for more
        sophisticated syntax in the filtering
        
        Usage:
        -------
        filters : Union[str, Iterable[str]]
            One or multiple patterns, which are OR'd together
        regex : bool, optional
            Accept plain regex syntax instead of shell-style, default: False
        
        Returns
        -------
        matches : pd.Series
        Mask for selecting matched rows
        """    
            
        # filter by columns and list of values
        keep = True

        for field, pattern in filters.items():
            # treat `col=None` as no filter applied
            if pattern is None:
                continue

            if field not in self.inventory:
                raise ValueError(f'filter by `{field}` not supported')

            keep &= util.pattern_match(
                self.inventory[field], pattern, regex=regex
            )

        if level is not None:
            keep &= self.inventory['variable'].str.count(r"\|") == level

        table = pd.DataFrame(
            self.inventory if keep is True else self.inventory.loc[keep]
        )

        # test to add function to a instance (does not require class)
        table.graph   = types.MethodType(util.plot_query_as_graph, table)
        table.short   = types.MethodType(util.shorten_find_output, table)
        table.unique  = types.MethodType(unique, table)
        
        table.filterp  = types.MethodType(util.filterp, table)
        table.load    = types.MethodType(self.getTables, table.index) 
        table.as_pyam = types.MethodType(util.to_pyam, table)
        table.as_xarray = types.MethodType(xarray.load_as_xdataset, 
                                           table)
        table.as_wide_dataframe    = types.MethodType(self.query_to_long_table, table) 
        return table
    
    
    def findExact(self, **kwargs):
        """
        Finds an exact match for the given filter criteria
        """
        table = self.inventory.copy()
        for key in kwargs.keys():
            mask = self.inventory[key] == kwargs[key]
            mask[pd.isna(mask)] = False
            mask = mask.astype(bool)
            table = table.loc[mask].copy()
        
        # test to add function to a instance (does not require class)
        table.graph = types.MethodType(util.plot_query_as_graph, table )
        table.short = types.MethodType(util.shorten_find_output, table)
        table.to_pyam = types.MethodType(util.to_pyam, table)
        return table
    
    def _getTableFilePath(self,ID):
        """
        Private
        Return the file path for a given tableID
        """
        source = self.inventory.loc[ID].source
        fileName = core.generate_table_file_name(ID)
        return os.path.join(config.PATH_TO_DATASHELF, 'database/', source, 'tables', fileName)

    # def _getTableFileName(self, ID):
    #     """
    #     For compatibility to windows based sytems, the pipe '|' symbols is replaces
    #     by double underscore '__' for the csv filename.
    #     """
    #     return ID.replace('|','-').replace('/','-') + '.csv'

    def _get_source_path(self, sourceID):
        return  os.path.join(config.PATH_TO_DATASHELF, 'database', sourceID)
    
    def getTable(self, ID, native_regions=False):
        """
        Method to load the datatable for the given tableID.
        
        Input
        -----
        tableID : str
        native_regions : bool, optional
            Load native region defintions if available. The default is False.

        Returns
        table : Datatable
        """
        if config.logTables:
            core.LOG['tableIDs'].append(ID)


        if self._tableExists(ID):
            # load table from database
            
            filePath = self._getTableFilePath(ID)
            
            table = read_csv(filePath,native_regions)
            #table = table[table.index.notnull()]
            return table
        
        elif os.path.exists('data'):
            fileName = core.generate_table_file_name(ID)
            filePath = os.path.join('data', fileName)
            
            if os.path.exists(filePath):
               table = read_csv(filePath,native_regions)
               #table = table[table.index.notnull()]
               return table
            
                
        
        if config.AUTOLOAD_SOURCES:
            # trying to import sources from remote on demand
            source = self._getSourceFromID(ID)
            
            if source in core.DB.sources.index:
                
                raise(BaseException('Table {} not found'.format(ID)))
            
            try: 
                if config.DEBUG:
                    print('Trying to import source {}'.format(source))
                self.importSourceFromRemote(source)
                    
                print("""###########################################################
                         Successfully imported source: {}
                         ###########################################################""".format(source))
                
                if self._tableExists(ID):
                # load table from database
                
                    filePath = self._getTableFilePath(ID)
                    table = read_csv(filePath,native_regions).drop_duplicates()
                    table = table[table.index.notnull()]
                    return table
            except Exception:
                
                if config.DEBUG:
                    print(traceback.format_exc())
                    print('Failed to import source {}'.format(source))
        
        raise(BaseException('Table {} not found'.format(ID)))
        
    def getTables(self, iterIDs, native_regions=False, disable_progress=None):
        """
        Method to return multiple datatables at once as a dictionary like 
        set fo tables.
        
        Input 
        -----
        iterIDS: list [str]
            List of IDs to load.
        native_regions : bool, optional
            Load native region defintions if available. The default is False.
        disable_progress : bool, optional
            Disable displaying of progressbar. The default None hides the
            progressbar on non-tty outputs

        Returns
        -------
        tables : TableSet
        """
        
        if config.logTables:
            IDs = list()
        res = TableSet()
        for ID in tqdm.tqdm(iterIDs, disable=disable_progress):
            table = self.getTable(ID,native_regions)
            if config.logTables:
                IDs.append(ID)
            res.add(table)
        if config.logTables:
            core.LOG['tableIDs'].extend(IDs)
        return res
    
    def getTablesAvailable(self):
        """
        Return a locally stored pandas dataframe of tables
        on datashelf
        """
        return pd.read_csv(os.path.join(config.MODULE_DATA_PATH, 'datashelf_contents.csv'))
  
    def startLogTables(self):
        """
        Starts the logging of loaded datatables. This is useful to collect all
        required tables for a given analysis to create a datapackage for off-line
        useage
        """
        config.logTables = True
        core.LOG['tableIDs'] = list()
    
    def stopLogTables(self):
        """ 
        Stops the logging process of datatables and return the list of loaded
        table IDs for more processing.
        """
        import copy
        config.logTables = False
        outList = copy.copy(core.LOG['tableIDs'])
        #core.LOG.TableList = list()
        return outList
    
    def clearLogTables(self):
        """ 
        Clears the list of logged tables. This is anyway done if the package is
        newly loaded
        """
        core.LOG['tableIDs'] = list()
    
    def save_logged_tables(self,
                           folder='data'):
        """
        Creates a local data directory that can be used to run
        the logges analysis indepenedly.
        

        Parameters
        ----------
        folder : str, optional
            DESCRIPTION. The default is 'data'.

        Returns
        -------
        None.

        """
        #create folder if required
        if  not os.path.exists(folder):
            os.mkdir(folder)
        #save tables to disk
        self.saveTablesToDisk(folder, core.LOG['tableIDs'])
        if config.DEBUG:
            print('{} tables stored to directory {}'.format(
                len(core.LOG['tableIDs']), folder))
        
        
    def saveTablesToDisk(self, folder, IDList):
        """ 
        Function to save a list of tables to disk as csv files.
        """
        from shutil import copyfile
        import os 

        for ID in IDList:
            pathToFile = self._getTableFilePath(ID)
            print()
            copyfile(pathToFile, folder + '/' + os.path.basename(pathToFile))
          
            
    def isSource(self, sourceID):
        """
        Checks is the source is in the database
        
        Input
        ------
        sourceID : str
        """
        return self.gitManager.isSource(sourceID)

    def commitTable(self, dataTable, message, sourceMetaDict=None):
        """
        Adds a table permamently to the underlying database. For the first table
        of a new source, the meta data for the sources needs to be provides as well
        
        Input
        ------
        table : Datatable
        message : str
        sourceMetaDict [Optional] :  dict
        """
        sourceID = dataTable.meta['source']
        if not self.isSource(sourceID):
            if sourceMetaDict is None:
                raise(BaseException('Source does not exist and no sourceMeta provided'))
            else:
                if not( sourceMetaDict['SOURCE_ID'] in self.gitManager.sources.index):
                    self._addNewSource(sourceMetaDict)
        
        dataTable = util.cleanDataTable(dataTable)
        self._addTable(dataTable)
        self.add_to_inventory(dataTable)
           
        self._gitCommit(message)

    def commitTables(self, 
                     dataTables, 
                     message, 
                     sourceMetaDict, 
                     append_data=False, 
                     update=False, 
                     overwrite=False , 
                     cleanTables=True):
        """
        Adds multipe tables permamently to the underlying database. For the first table
        of a new source, the meta data for the sources needs to be provides as well
        
        Input
        ------
        tables : list of Datatable
        message : str
        sourceMetaDict [Optional] :  dict
        append_data [optinal]  : bool to choose if new data is added to the existing
                                 table (new data does not overwrite old data)
        update : [optional]    : bool to choose if the exting data is updated
        overwrire : [optional] : bool to choose if data is overwriten (new data 
                                 overwrites old data)
        cleanTables [optional] : bool (default: true) to choose if tables are 
                                 cleaned before commit
        
        TODO: Check flags
        """
        # create a new source if not extisting
        if not self.isSource(sourceMetaDict['SOURCE_ID']):
            self._addNewSource(sourceMetaDict)

        # only test if an table is update if the source did exist
        if update:
            oldTableIDs = [table.generateTableID() for table in dataTables]
            self.updateTables(oldTableIDs, dataTables, message)
            return            
        
        else:
            for dataTable in tqdm.tqdm(dataTables):
                
            
                if config.DEBUG:
                    print(dataTable.ID)
                if cleanTables:
                    dataTable = util.cleanDataTable(dataTable)
                
                if dataTable.isnull().all().all():
                    print('ommiting empty table: ' + dataTable.ID)
                    continue
                
                if dataTable.ID not in self.inventory.index:
                    # add entire new table
                    
                    self._addTable(dataTable)
                    self.add_to_inventory(dataTable)
                elif overwrite:
                    self._addTable(dataTable)
                elif append_data:
                    # append data to table
                    oldTable = self.getTable(dataTable.ID, native_regions=True)
                    mergedTable = oldTable.combine_first(dataTable)
                    mergedTable = Datatable(mergedTable, meta = dataTable.meta)
                    self._addTable(mergedTable)
                
        self._gitCommit(message)
        
    def update_mapping_file(self, 
                            sourceID, 
                            mapping_file_path,
                            sourceMetaDict = None):
        """
        adds mapping file to database

        Parameters
        ----------
        source : TYPE
            DESCRIPTION.
        ID : TYPE
            DESCRIPTION.
        mapping : TYPE
            DESCRIPTION.
            """
        if not core.DB.isSource(sourceMetaDict['SOURCE_ID']):
            if sourceMetaDict is None:
                raise(BaseException('Source meta can not be None'))
            core.DB._addNewSource(sourceMetaDict)

        source_path = self._get_source_path(sourceID)
        # for key in mapping.keys():
        dest_file_path = os.path.join(source_path, 'mapping.xlsx')
        shutil.copyfile(mapping_file_path, dest_file_path)
            # mapping[key].to_csv(file_path)
        
        self.gitManager.gitAddFile(sourceID, dest_file_path)
        
    def updateTable(self, oldTableID, newDataTable, message):
        """
        Specific method to update the data of an existing table
        
        Input
        -----
        oldTableID    : str
        newDataTabble : Datatable
        message       : str 
                        Commit message to describle the added data
        """
        sourceID = self._getSourceFromID(newDataTable.ID)
        if not self.isSource(sourceID):
            raise(BaseException('source  does not exist'))
            
        self._updateTable(oldTableID, newDataTable)
        self._gitCommit(message)
    
    def updateTables(self, oldTableIDs, newDataTables, message):
        """
        Equivalent method to updateTable, but for multiple tables at once
        
        Input
        -----
        oldTableIDs    : list of str
        newDataTabbles : list of Datatable
        message        : str 
                         Commit message to describle the added data
        """ 
        
        
        sourcesToUpdate = list()
        for tableID in oldTableIDs:
            sourceID = self._getSourceFromID(tableID)
            if sourceID not in sourcesToUpdate:
                sourcesToUpdate.append(sourceID)
        
        # check that all sources do exist
        for sourceID in sourcesToUpdate:
            if not self.isSource(sourceID):
                raise(BaseException('source  does not exist'))
        
        # loop over tables
        for oldTableID, newDataTable in tqdm.tqdm(zip(oldTableIDs, newDataTables), desc= 'Updating tables'):
            
            if oldTableID in self.inventory.index:
                self._updateTable(oldTableID, newDataTable)
                
            else:
                dataTable = util.cleanDataTable(newDataTable)
                self._addTable(dataTable)
                self.add_to_inventory(dataTable)
        self._gitCommit(message)

    def _updateTable(self, oldTableID, newDataTable):
        """
        Private
        
        Method as a common function form multiple functions
        """
        newDataTable = util.cleanDataTable(newDataTable)
        
        oldDataTable = self.getTable(oldTableID)
        oldID = oldDataTable.meta['ID']
        newID = newDataTable.generateTableID()
        
        if oldID == newID:# and (oldDataTable.meta['unit'] == newDataTable.meta['unit']):
            #only change data
            self._addTable( newDataTable)
        else:
            # delete old table
            self.removeTable(oldID)
            
            # add new table
            self._addTable( newDataTable)

            #change inventory
            self.inventory.rename(index = {oldID: newID}, inplace = True)
        self.add_to_inventory(newDataTable)


    def updateTablesAvailable(self,private_access_token):
        """
        Method to update module data folder with the latest 
        datashelf contents on Gitlab

        Requirements:
        
            - private_access_token: generated on Gitlab

        """

        try:
            import gitlab
        except ImportError:
            print('python-gitlab not available: Install with mamba or conda')
            raise

        gl = gitlab.Gitlab(private_token=private_access_token)
        # define group 
        group = gl.groups.get("climateanalytics/datashelf")
        projects = group.projects.list(include_subgroups=True, all=True)
        print('fetching datashelf contents...')
        # init list
        rows = []
        for project in projects:
            # append
            rows.append({'table'          : project.name, 
                         'last_update'    : pd.to_datetime(project.last_activity_at),
                         'description'    : project.description,
                         #'fetched'        : pd.to_datetime('today')
                         })
        # make df
        tables = pd.DataFrame(rows)
        # change datetime
        tables['last_update'] = tables['last_update'].dt.strftime('%d-%m-%Y')
        #tables['fetched']     = tables['fetched'].dt.strftime('%d-%m-%Y')
        # overwrite table in module data
        tables.to_csv(os.path.join(config.MODULE_DATA_PATH, 'datashelf_contents.csv'),index=False)
        print('done')


    def validate_ID(self, ID, print_statement=True):
        """ 
        Method to chekc the validity of a table ID and check the state of the 
        data
        """
        RED = '\033[31m'
        GREEN = '\033[32m'
        BLACK = '\033[30m'
        source = ID.split(config.ID_SEPARATOR)[-1]
        print('TableID: {}'.format(ID))
        valid = list()
        if self.sourceExists(source):
            if print_statement:
                print(GREEN + "Source {} does exists".format(source))
            valid.append(True)
        else:
            if print_statement:
                print(RED + "Source {} does not exists".format(source))
            valid.append(False)
        if ID in self.inventory.index:
            if print_statement:
                print(GREEN + "ID is in the inventory")
            valid.append(True)
        else:
            if print_statement:
                print(RED + "ID is missing in the inventory")
            valid.append(False)
            
        fileName = core.generate_table_file_name(ID)
        tablePath = os.path.join(config.PATH_TO_DATASHELF, 'database', source, 'tables', fileName)

        if os.path.isfile(tablePath):
            if print_statement:
                print(GREEN + "csv file exists")
            valid.append(True)
        else:
            if print_statement:
                print(RED + "csv file does not exists")
            
            valid.append(False)

        print(BLACK)

        return all(valid)

    def removeTables_of_source(self, sourceID, IDList):
        
        if not self.isSource(sourceID):
            raise(BaseException('source  does not exist'))
            
        files_on_disk = list()
        for tableID in IDList:
            filePath = core.DB._getTableFilePath(tableID)
            if os.path.exists(filePath):
                files_on_disk.append(tableID)
        
        files = ['tables/' + core.generate_table_file_name(x) for x in files_on_disk]
        
        self.gitManager.gitRemoveFiles(sourceID,files)
        
        self.remove_from_inventory(IDList)
        
        self._gitCommit(f'{len(IDList)} tables of {sourceID} removed')
        
    def removeTables(self, IDList):
        """
        Method to remnove tables from the database
        
        Input
        -----
        IDList : list of str
        """
        
        sourcesToUpdate = list()
        for tableID in IDList:
            sourceID = self._getSourceFromID(tableID)
            if sourceID not in sourcesToUpdate:
                sourcesToUpdate.append(sourceID)
        
        # check that all sources do exist
        for source in sourcesToUpdate:
            if not self.isSource(sourceID):
                raise(BaseException('source  does not exist'))
        
        for ID in tqdm.tqdm(IDList,desc="Removing tables"):

            source = self.inventory.loc[ID, 'source']
            tablePath = self._getTableFilePath(ID)
            self.gitManager.gitRemoveFile(source, tablePath)
            
        self.remove_from_inventory(IDList)

        self._gitCommit('Tables removed')
        
    def removeTable(self, tableID):
        """
        Method to remnove tables from the database
        
        Input
        -----
        tableID : str
        """
        sourceID = self._getSourceFromID(tableID)
        if not self.isSource(sourceID):
            raise(BaseException('source  does not exist'))
        
        self._removeTable(tableID)
        self._gitCommit('Table removed')
        self._reloadInventory()

    def _removeTable(self, ID):
        """
        Private method
        Function to pool code for removing a table from the database
        """
        source = self.inventory.loc[ID, 'source']
        tablePath = self._getTableFilePath(ID)
        self.remove_from_inventory(ID)
#        self.gitManager[source].execute(["git", "rm", tablePath])
        self.gitManager.gitRemoveFile(source, tablePath)

        
        
    def _tableExists(self, ID):
        return ID in self.inventory.index
    


    def tableExist(self, tableID):
        """
        Method to test if table exists in the database
        
        Input
        -----
        tableID : str
        """
        
        return self._tableExists(tableID)

    def isConsistentTable(self, datatable):
        """
        Checks if that table is fitting the following requirements 
        - numeric data 
        - spatial identifiers are known to the database
        - columns are propper years
        - index is not duplicated
        """
        if not np.issubdtype(datatable.values.dtype, np.number):
            
            raise(BaseException('Sorry, data of table {} is needed to be numeric'.format(datatable)))            
            
        # check that spatial index is consistend with defined countries or regions
        # invalidSpatialIDs = datatable.index.difference(mapp.getValidSpatialIDs())
        # if len(invalidSpatialIDs) > 0:
        #     raise(BaseException('Sorry, regions in table {}: {} do not exist'.format(datatable, invalidSpatialIDs)))
        
        # check that the time colmns are years
        from pandas.api.types import is_integer_dtype
        if not is_integer_dtype(datatable.columns):
            raise(BaseException('Sorry, year index in table {} is not integer'.format(datatable)))

        if sum(datatable.index.duplicated()) > 0:
            print(datatable.meta)
            raise(BaseException('Sorry, region index in table {} is not  unique'.format(datatable)))
        return True
    

    def _addTable(self, datatable):
        """
        Private
        Pools functionality to add table to the database
        """
        
        # ID = datatable.generateTableID()
        source = datatable.source()
        datatable.meta['creator'] = config.CRUNCHER
        sourcePath = os.path.join('database', source)
        filePath = os.path.join(sourcePath, 'tables',  datatable.getTableFileName())
        if (config.OS == 'win32') | (config.OS == "Windows"):
            filePath = filePath.replace('|','___')
        
        if 'standard_region' in datatable.columns or 'region' in datatable.columns:
            datatable = datatable.reset_index().set_index(['region', 'standard_region'])
            datatable.columns = datatable.columns.astype(int)
        
        datatable = util.cleanDataTable(datatable)
        datatable = datatable.sort_index(axis='index')
        datatable = datatable.sort_index(axis='columns')
        
        
        
        self.isConsistentTable(datatable)
        

        self._gitAddTable(datatable, source, filePath)

    def _gitAddTable(self, datatable, source, filePath):
        """
        Private
        Added file to git system
        """
        datatable.to_csv(os.path.join(config.PATH_TO_DATASHELF, filePath))
        ID = datatable.ID
        self.gitManager.gitAddFile(source, os.path.join('tables', core.generate_table_file_name(ID)))
        
    def _gitCommit(self, message):
        """
        Private
        Commits all changes to git
        """
        self.inventory.to_csv(self.INVENTORY_PATH)
        self.gitManager.gitAddFile('main',self.INVENTORY_PATH)

        for sourceID in self.gitManager.updatedRepos:
            if sourceID == 'main':
                continue
            repoPath = os.path.join(config.PATH_TO_DATASHELF, 'database', sourceID)
            sourceInventory = self.inventory.loc[self.inventory.source==sourceID,:]
            sourceInventory.to_csv(os.path.join(repoPath, 'source_inventory.csv'))
            self.gitManager.gitAddFile(sourceID, os.path.join(repoPath, 'source_inventory.csv'))  
        self.gitManager.commit(message)
            

    def _addNewSource(self, sourceMetaDict):
        """
        Private
        Adds new source to the sources table 
        """
        source_ID = sourceMetaDict['SOURCE_ID']
        
        if not self.sourceExists(source_ID):
            
            sourcePath = os.path.join(config.PATH_TO_DATASHELF, 'database', sourceMetaDict['SOURCE_ID'])
            self.gitManager.init_new_repo(sourcePath, source_ID, sourceMetaDict)
            

        else:
            print('source already exists')

    def _getSourceFromID(self, tableID):
        """
        Private
        Returns the source of a given table
        """
        return tableID.split(config.ID_SEPARATOR)[-1]
    
    
    def removeSource(self, sourceID):
        """
        Function to remove an entire source from the database. 
        """
        
        if self.sourceExists(sourceID):
            sourcePath = os.path.join(config.PATH_TO_DATASHELF, 'database', sourceID)
            shutil.rmtree(sourcePath, ignore_errors=False, onerror=None)
        self.gitManager.sources = self.gitManager.sources.drop(sourceID, axis=0)
        self.sources            = self.gitManager.sources
        self.inventory = self.inventory.loc[self.inventory.source != sourceID]
#        self.gitManager.updatedRepos.add('main')
        self._gitCommit(sourceID + ' deleted')
    
    def updateExcelInput(self, fileName):
        """
        This function updates all data values that are defined in the input sheet
        in the given excel file
        """
        if config.DB_READ_ONLY:
            assert self._validateRepository()
        ins = io.Inserter(fileName='demo.xlsx')
        for setup in ins.getSetups():
            dataTable = self.getTable(setup['dataID'])
            ins._writeData(setup, dataTable)
            
    
    #database mangement
    
    def _checkTablesOnDisk(self, sourceID=None):
        notExistingTables = list()
        
        if sourceID is None:
            table_list = self.inventory.index
        else:
            table_list = self.inventory.index[self.inventory.source == sourceID]
        
        for tableID in self.inventory.index:
            filePath = self._getTableFilePath(tableID)
            if not os.path.exists(filePath):
                notExistingTables.append(tableID)
        
        return notExistingTables
                
      
    def importSourceFromRemote(self, remoteName):
        """
        This functions imports (git clone) a remote dataset and creates a local
        copy of it.
        
        Input is an existing sourceID. 
        """
        repoPath = os.path.join(config.PATH_TO_DATASHELF, 'database', remoteName)
        
        self.gitManager.clone_source_from_remote(remoteName, repoPath)

        sourceInventory = pd.read_csv(os.path.join(repoPath, 'source_inventory.csv'), index_col=0, dtype={'source_year': str})
        self.inventory = self.inventory.append(sourceInventory, verify_integrity=True)
        self._gitCommit('imported ' + remoteName)

    def exportSourceToRemote(self, sourceID):
        """
        This function exports a new local dataset to the defind remote database.
        
        Input is a local sourceID as a str.
        """
        self.gitManager.create_remote_repo(sourceID)
        self.gitManager.push_to_remote_datashelf(sourceID)
        print('export successful: ({})'.format( config.DATASHELF_REMOTE +  sourceID))
        

    def pull_update_from_remote(self, repoName):
        """
        Updates the local data repository by the newest version on the remote repository
        

        Parameters
        ----------
        repoName : str
            Source ID string to identify which source repository should be updated.

        Returns
        -------
        None.

        """
        new_inventory = self.gitManager.pull_update_from_remote(repoName,self.inventory)
        self.inventory = new_inventory
        
        self._gitCommit('udpate from remote')
        
    
    def query_to_long_table(self, 
                            query, 
                            native_regions=False,
                            meta_list = ['variable', 'region','scenario', 'model', 'source', 'unit']):
    
    
        tables = [self.getTable(x) for x in query.index]
        final_tables = list()
        for table in tables:
            if table.empty:
                continue
            
            if native_regions:
                table = table.reset_index('standard_region',drop=True)
            else:
                table = table.reset_index('region',drop=True)
                
            inp_dict = dict()    
            for metaKey in meta_list:
                if metaKey == 'region':
                    inp_dict[metaKey] = table.index
                else:
                    inp_dict[metaKey] = table.meta.get(metaKey, '')
                
            try:
                final_tables.append(pd.DataFrame(table.assign(**inp_dict)).reset_index(drop=True))
            except KeyError as exc:
                raise AssertionError(f"meta of {table.ID} does not contain {exc.args[0]}")
        long_df = pd.concat(final_tables, ignore_index=True, sort=False)
        
        # move id columns to the front
        id_cols = pd.Index(meta_list)
        long_df = long_df[list(id_cols) + list( long_df.columns.difference(id_cols))]
        long_df = pd.DataFrame(long_df)
        return long_df       
#%%
class GitRepository_Manager:
    """
    # Management of git repositories for fast access
    """
    def __init__(self, 
                 config,
                 debugmode = False):
        self.PATH_TO_DATASHELF = config.PATH_TO_DATASHELF
        self.sources = pd.read_csv(config.SOURCE_FILE, index_col='SOURCE_ID')
        
        self.repositories = dict()
        self.updatedRepos = set()
        self.validatedRepos = set()
        self.filesToAdd   = defaultdict(list)
        
        if not debugmode:
            for sourceID in self.sources.index:
                repoPath = os.path.join(self.PATH_TO_DATASHELF,  'database', sourceID)
                self.repositories[sourceID] = git.Repo(repoPath)
                self.verifyGitHash(sourceID)
            
            self.repositories['main'] = git.Repo(self.PATH_TO_DATASHELF)
            self._validateRepository('main')
        else:
            print('Git manager initialized in debugmode')
            
    def get_source_repo_failsave(self, sourceID):
        """
        Retrieve `sourceID` from repositories dictionary without checks
        """
        repoPath = os.path.join(self.PATH_TO_DATASHELF,  'database', sourceID)
        repo = git.Repo(repoPath)
        return repo
    
    def __getitem__(self, sourceID):
        """ 
        Retrieve `sourceID` from repositories dictionary and ensure cleanliness
        """
        repo = self.repositories[sourceID]
        if sourceID not in self.validatedRepos:
            self._validateRepository(sourceID)
        return repo
    
    def _validateRepository(self, sourceID):
        """ 
        Private
        Cheks if sourceID points to a valid repository
        
        """
        repo = self.repositories[sourceID]

        if sourceID != 'main':
            self.verifyGitHash(sourceID)

        if repo.is_dirty():
            raise RuntimeError('Git repo: "{}" is inconsistent! - please check uncommitted modifications'.format(sourceID))

        config.DB_READ_ONLY = False
        if config.DEBUG:
            print('Repo {} is clean'.format(sourceID))
        self.validatedRepos.add(sourceID)
        return True
        
    def init_new_repo(self, repoPath, repoID, sourceMetaDict):
        """
        Method to create a new repository for a source
        
        Input
        ----
        repoPath : str
        repoID   : str 
        sourceMetaDict : dict with the required meta data descriptors
        """
        self.sources.loc[repoID] = pd.Series(sourceMetaDict)
        self.sources.to_csv(config.SOURCE_FILE)
        self.gitAddFile('main', config.SOURCE_FILE)

        repoPath = Path(repoPath)
        print(f'creating folder {repoPath}')
        repoPath.mkdir(parents=True, exist_ok=True)
        self.repositories[repoID] = git.Repo.init(repoPath)

        for subFolder in config.SOURCE_SUB_FOLDERS:
            subPath = repoPath / subFolder
            subPath.mkdir(exist_ok=True)
            filePath = subPath / '.gitkeep'
            filePath.touch()
            self.gitAddFile(repoID, filePath)

        metaFilePath = repoPath / 'meta.csv'
        util.dict_to_csv(sourceMetaDict, metaFilePath)
        self.gitAddFile(repoID, metaFilePath)

        self.commit('added source: ' + repoID)


    def gitAddFile(self, repoName, filePath):
        """
        Addes a new file to a repository
        
        Input
        -----
        repoName : str 
        filePath : str of the relative file path
        """
        if config.DEBUG:
            print('Added file {} to repo: {}'.format(filePath,repoName))
        
        self.filesToAdd[repoName].append(str(filePath))
        self.updatedRepos.add(repoName)
        
    def gitRemoveFiles(self, repoName, filePaths):
        """
        Removes mutiple file from the git repository
        
        Input
        -----
        repoName : str 
        filePath : str of the relative file path
        """
        self[repoName].index.remove(filePaths, working_tree=True)
        self.updatedRepos.add(repoName)
        
    def gitRemoveFile(self, repoName, filePath):
        """
        Removes a file from the git repository
        
        Input
        -----
        repoName : str 
        filePath : str of the relative file path
        """
        if config.DEBUG:
            print('Removed file {} to repo: {}'.format(filePath,repoName))
        self[repoName].git.execute(["git", "rm", "-f", filePath]) #TODO Only works with -f forced, but why?
        self.updatedRepos.add(repoName)
    
    def _gitUpdateFile(self, repoName, filePath):
        pass
        
    def commit(self, message):
        """
        Function to commit all oustanding changes to git. All repos including 
        'main' are commited if there is any change
        
        Input
        ----
        message : str - commit message
        """
        if 'main' in self.updatedRepos:
            self.updatedRepos.remove('main')

        for repoID in self.updatedRepos:
            repo = self.repositories[repoID]
            repo.index.add(self.filesToAdd[repoID])
            commit = repo.index.commit(message + " by " + config.CRUNCHER)
            self.sources.loc[repoID, 'git_commit_hash'] = commit.hexsha
            del self.filesToAdd[repoID]
        
        # commit main repository
        self.sources.to_csv(config.SOURCE_FILE)
        self.gitAddFile('main', config.SOURCE_FILE)

        main_repo = self['main']
        main_repo.index.add(self.filesToAdd['main'])
        main_repo.index.commit(message + " by " + config.CRUNCHER)
        del self.filesToAdd['main']

        #reset updated repos to empty
        self.updatedRepos        = set()
        
        
    def create_remote_repo(self, repoName):
        """
        Function to create a remote git repository from an existing local repo
        """
        repo = self[repoName]

        if 'origin' in repo.remotes:
            # remote origin has been configured already, but re-push anyway, since this
            # could have been a connectivity issue
            origin = repo.remotes.origin
        else:
            origin = repo.create_remote("origin", config.DATASHELF_REMOTE + repoName + ".git")

        branch = repo.active_branch
        origin.push(branch, progress=TqdmProgressPrinter())

        # Update references on remote
        origin.fetch()
        branch.set_tracking_branch(origin.refs[0])
    
    def push_to_remote_datashelf(self, repoName):
        """
        This function used git push to update the remote database with an updated
        source dataset. 
        
        Input is the source ID as a str.
        
        Currently conflicts beyond auto-conflict management are not caught by this
        function. TODO

        """
        self[repoName].remotes.origin.push(progress=TqdmProgressPrinter())
        
    def clone_source_from_remote(self, repoName, repoPath):
        """
        Function to clone a remote git repository as a local copy.
        
        Input
        ----- 
        repoName : str - valid repository in the remove database
        repoPath : str - path of the repository
        """
        try:
            print('Try cloning source via ssh')
            url = config.DATASHELF_REMOTE + repoName + '.git'
            repo = git.Repo.clone_from(url=url, to_path=repoPath, progress=TqdmProgressPrinter())  
        except:
            print('Failed... Try Cloning source via https:')
            url = config.DATASHELF_REMOTE_HTTPS + repoName + '.git'
            repo = git.Repo.clone_from(url=url, to_path=repoPath, progress=TqdmProgressPrinter())

        self.repositories[repoName] = repo

        # Update source file
        sourceMetaDict = util.csv_to_dict(os.path.join(repoPath, 'meta.csv'))
        sourceMetaDict['git_commit_hash'] = repo.commit().hexsha
        self.sources.loc[repoName] = pd.Series(sourceMetaDict)   
        self.sources.to_csv(config.SOURCE_FILE)
        self.gitAddFile('main', config.SOURCE_FILE) 

        return repo
        
    def pull_update_from_remote(self, repoName, old_inventory):
        """
        This function used git pull an updated remote source dataset to the local
        database.
        
        Input is the source ID as a str.
        
        Currently conflicts beyond auto-conflict management are not caught by this
        function. TODO

        """
        self[repoName].remote('origin').pull(progress=TqdmProgressPrinter())
        self.updateGitHash(repoName)
        repoPath = os.path.join(self.PATH_TO_DATASHELF,  'database', repoName)
        sourceInventory = pd.read_csv(os.path.join(repoPath, 'source_inventory.csv'), index_col=0, dtype={'source_year': str})
        new_inventory = pd.concat([
            old_inventory[old_inventory["source"] != repoName],
            sourceInventory
        ])

        
        return new_inventory
    
    def verifyGitHash(self, repoName):
        """
        Function to verify the git hash code of an existing git repository
        """
        repo = self.repositories[repoName]
        if repo.commit().hexsha != self.sources.loc[repoName, 'git_commit_hash']:
            raise RuntimeError('Source {} is inconsistent with overall database'.format(repoName))

    def updateGitHash(self, repoName):
        """
        Function to update the git hash code in the sources.csv by the repo hash code
        """
        self.sources.loc[repoName,'git_commit_hash'] = self[repoName].commit().hexsha
        
    def setActive(self, repoName):
        """
        Function to set a reposity active
        """
        self[repoName].git.refresh()
        
    def isSource(self, sourceID):
        if sourceID in self.sources.index:
            self[sourceID].git.refresh()
            return True
        else:
            return False
        
class TqdmProgressPrinter(git.RemoteProgress):
    known_ops = {
        git.RemoteProgress.COUNTING: "counting objects",
        git.RemoteProgress.COMPRESSING: "compressing objects",
        git.RemoteProgress.WRITING: "writing objects",
        git.RemoteProgress.RECEIVING: "receiving objects",
        git.RemoteProgress.RESOLVING: "resolving stuff",
        git.RemoteProgress.FINDING_SOURCES: "finding sources",
        git.RemoteProgress.CHECKING_OUT: "checking things out"
    }

    def __init__(self):
        super().__init__()
        self.progressbar = None

    def update(self, op_code, cur_count, max_count=None, message=''):
        if op_code & self.BEGIN:
            desc = self.known_ops.get(op_code & self.OP_MASK)
            self.progressbar = tqdm.tqdm(desc=desc, total=max_count)

        self.progressbar.set_postfix_str(message, refresh=False)
        self.progressbar.update(cur_count)

        if op_code & self.END:
            self.progressbar.close()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains all relevant tools regarding the use of excel


@author: ageiges
"""
import os
import re
import string
import pandas as pd
import tqdm
import platform

from openpyxl import load_workbook


#%% Defintions
OS = platform.system() #'win32' , linux, #Darwin


REQUIRED_SETUP_FIELDS = ['filePath',
                         'fileName',
                         'sheetName',
                         'timeIdxList',
                         'spaceIdxList']

SP_ARG = 3
TI_ARG = 2
DT_ARG = 4

REG_EXCEL_RANGE = re.compile('^[A-Z]{1,3}[0-9]{1,3}:[A-Z]{1,3}[0-9]{1,3}$')
REG_EXCEL_ROW   = re.compile('^[0-9]{1,2}$')
REG_EXCEL_COL   = re.compile('^[A-Z]{1,2}$')

REG_FIND_ROWS   = re.compile('^[A-Z]{1,3}([0-9]{1,3}):[A-Z]{1,3}([0-9]{1,3})$')
REG_FIND_COLS   = re.compile('^([A-Z]{1,3})[0-9]{1,3}:([A-Z]{1,3})[0-9]{1,3}$')





    

#%% Private functions
def _isColRange(string):
#    if REG_EXCEL_COL.match(string):
#        return True
    match = REG_FIND_COLS.search(string)
    if match:
        if match.group(1) == match.group(2):
            return True
    
    return False
        
def _isRowRange(string):
#    if REG_EXCEL_ROW.match(string):
#        return True
    match = REG_FIND_ROWS.search(string)
    if match:
        if match.group(1) == match.group(2):
            return True
    
    return False

def _isRow(string):
    if REG_EXCEL_ROW.match(string):
        return True
    else:
        return False
    
def _isCol(string):
    if REG_EXCEL_COL.match(string):
        return True
    else:
        return False
#%%
def _iterTime(timeIdxString, wksheet, xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx):
    if _isRowRange(timeIdxString):
        # colum setup
        for timeCell in wksheet[timeIdxString][0]:
            xlsCol = timeCell.col_idx
            timeIdx = timeCell.value
            #print(xlsCol, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    
    elif _isRow(timeIdxString):
        for timeCell in wksheet[timeIdxString]:
            xlsCol = timeCell.col_idx
            timeIdx = timeCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    elif _isCol(timeIdxString):
        for timeCell in wksheet[timeIdxString]:
            xlsRow = timeCell.row
            timeIdx = timeCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
            
    elif _isColRange(timeIdxString):
        for timeCell in wksheet[timeIdxString]:
#            print(timeCell[0].row)
            xlsRow = timeCell[0].row
            timeIdx = timeCell[0].value
            #print(xlsRow, timeIdx)
#            print([xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx])
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    else:
        # assume fixed
        timeIdx = timeIdxString
        yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]

def _iterSpace(spaceIdxString, wksheet, xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx):
    if _isRowRange(spaceIdxString):
        # colum setup
        for spaceCell in wksheet[spaceIdxString][0]:
            xlsCol = spaceCell.col_idx
            spaceIdx = spaceCell.value
            #print(xlsCol, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]

    elif _isRow(spaceIdxString):
        for spaceCell in wksheet[spaceIdxString]:
            xlsCol = spaceCell.col_idx
            spaceIdx = spaceCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    
    elif _isCol(spaceIdxString):
        for spaceCell in wksheet[spaceIdxString]:
            xlsRow = spaceCell.row
            spaceIdx = spaceCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
            
    elif _isColRange(spaceIdxString):
        for spaceCell in wksheet[spaceIdxString]:
            xlsRow = spaceCell[0].row
            spaceIdx = spaceCell[0].value
            #print(xlsRow, spaceIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    else:
        # assume fixed string
        print('fixed')
        spaceIdx = spaceIdxString
        yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
        
def _iterData(dataIdxString, wksheet, xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx):

    #print([xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx])
    if _isRowRange(dataIdxString):
        # colum setup
        for dataCell in wksheet[dataIdxString][0]:
            xlsCol = dataCell.col_idx
            dataIdx = dataCell.value
            #print(xlsCol, timeIdx, dataIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]

    elif _isRow(dataIdxString):
        for dataCell in wksheet[dataIdxString]:
            xlsCol = dataCell.col_idx
            dataIdx = dataCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    
    
    elif _isCol(dataIdxString):
#        print('_isCol')
#        print([xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx])
    
        for i, dataCell in enumerate(wksheet[dataIdxString]):
#            print(dataCell)
#            print(dataCell.col_idx)
#            xlsRow = dataCell.row
            xlsCol = dataCell.col_idx
            if i == 0:
                dataIdx = dataCell.value
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
            
    elif _isColRange(dataIdxString):
#        print('_isColRange')
        for dataCell in wksheet[dataIdxString]:
            #print(dataCell)
            xlsRow = dataCell[0].row
            if dataCell[0].value is not None:
                dataIdx = dataCell[0].value.replace('\ufeff','')
            else:
                dataIdx = None
            #print(xlsRow, timeIdx)
            yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]
    else:
        # assume fixed
        dataIdx = dataIdxString
        yield  [xlsRow, xlsCol, timeIdx, spaceIdx, dataIdx]

def _identifyCountry(string):
    """
    Used various methods to find best fitting country alpha3 iso code
    """
    
    import datatoolbox as dt
    #numeric ISO code
    try:
        numISO = float(string)
        mask = numISO == dt.mapp.countries.codes['numISO']
        if mask.any():
            return dt.mapp.countries.index[mask][0]
    except:
        pass
    if len(string) == 2:
        mask = string == dt.mapp.countries.codes['alpha2']
        if mask.any():
            return dt.mapp.countries.codes.index[mask][0]
    
    if len(string) == 3:
         
        if string.upper() in dt.mapp.countries.codes.index:
            return string.upper()
    
    try: 
        coISO = dt.getCountryISO(string)
        return coISO
    except:
        print('not matching country found')    
        return None

def _res2Excel(resFull, countryList):
    import datatoolbox as dt
    tableSet = dt.TableSet()
    
    yearRange = (pd.np.inf, -pd.np.inf)
    
    ID_list = list()
    for ID in resFull:
        table = dt.getTable(ID)
        if 'Gg' in table.meta['unit']:
            table = table.convert(table.meta['unit'].replace('Gg','Mt'))
        tableSet.add(table)
        ID_list.append(table.ID)
        
        minYears = min(yearRange[0], table.columns.min())
        maxYears = max(yearRange[1], table.columns.max())
        yearRange = (minYears, maxYears)
        
    for country in countryList:
        
        coISO = _identifyCountry(country)
        
        if coISO is None:
            coISO = country
        
        outDf = pd.DataFrame(columns = ['Source', 'Entity', 'Category', 'Scenario', 'Unit'] + list(range(yearRange[0], yearRange[1]+1)))
        
        i =0
        for ID in ID_list:
            if coISO in tableSet[ID].index:
                years =  tableSet[ID].columns[~tableSet[ID].loc[coISO].isna()]
                outDf.loc[i,years] = tableSet[ID].loc[coISO,years]
                outDf.loc[i,['Source', 'Entity','Category', 'Scenario', 'Unit']] = [tableSet[ID].meta[x] for x in ['source', 'entity','category', 'scenario', 'unit']]
                i = i+1
        
        outDf = outDf.loc[:,~outDf.isnull().all(axis=0)]
        outDf.to_excel('extract_'+  coISO + '.xlsx')
 
def _str2float(x):
    if isinstance(x,float) or isinstance(x,int) or x is None:
        return x
    if '%' in x:
        return float(x.replace('%',''))*100
    else:
        if x.startswith('#') or x == '':
            return np.nan
#        try:
        print(x)
        return float(x)          
#%% Functions
        
def alphaCol2Num(col):
    num = 0
    for c in col:
        if c in string.ascii_letters:
            num = num * 26 + (ord(c.upper()) - ord('A')) +1 
    return num-1

def colNameToNum(name):
    pow = 1
    colNum = 0
    for letter in name[::-1]:
            colNum += (int(letter, 36) -9) * pow
            pow *= 26
    return colNum -1

def compare_excel_files(file1, file2, eps=1e-6):
    """
    This function compares two excel files and creates a file visualising the differences
    
    """
    
    def report_diff(x):
        try:
#            print(x)
            x = x.astype(float)
#            print(x)
            return x[1] if (abs(x[0] - x[1]) < eps) or  pd.np.any(pd.isnull(x))  else '{0:.2f} -> {1:.2f}'.format(*x)
        except:
            return x[1] if (x[0] == x[1]) or  pd.np.any(pd.isnull(x))  else '{} -> {}'.format(*x)
            #print(x)
            
    xlFile1 = pd.ExcelFile(file1)
    sheetNameList1 = set(xlFile1.sheet_names)
    
    xlFile2 = pd.ExcelFile(file2)
    sheetNameList2 = set(xlFile2.sheet_names)
    
    writer = pd.ExcelWriter("diff_temp.xlsx")
    
    for sheetName in sheetNameList2.intersection(sheetNameList1):
        data1  = pd.read_excel(xlFile1, sheet_name=sheetName)
        data2  = pd.read_excel(xlFile2, sheet_name=sheetName)
        data1 = data1.replace(pd.np.nan, '').astype(str)
        data2 = data2.replace(pd.np.nan, '').astype(str)
        data1 = data1.apply(lambda x: x.str.strip())
        data2 = data2.apply(lambda x: x.str.strip())
        diff_panel = pd.Panel(dict(df1=data1,df2=data2))
        diff_output = diff_panel.apply(report_diff, axis=0)

        diff_output.to_excel(writer,sheet_name=sheetName)
    
        workbook  = writer.book
        worksheet = writer.sheets[sheetName]
        highlight_fmt = workbook.add_format({'font_color': '#FF0000', 'bg_color':'#B1B3B3'})
    
        grey_fmt = workbook.add_format({'font_color': '#8d8f91'})
        
        worksheet.conditional_format('A1:ZZ1000', {'type': 'text',
                                                'criteria': 'containing',
                                                'value':'->',
                                                'format': highlight_fmt})
        worksheet.conditional_format('A1:ZZ1000', {'type': 'text',
                                                'criteria': 'not containing',
                                                'value':'->',
                                                'format': grey_fmt})
        
    writer.close()
    os.system('libreoffice ' + "diff_temp.xlsx")
    os.system('rm diff_temp.xlsx')


def excel_to_pandas_idx(index):
    """
    Conversion from Excel Row-Col sting to pandas index
    
    E.g. excel_to_pandas_idx('B3')

    Parameters
    ----------
    index : str
        Excel index

    Returns
    -------
    tuple coordinate index

    """
    [alpha, num, _] = re.split(r'(\d+)',index)    
    return (int(num)-1, alphaCol2Num(alpha))

def getAllSheetNames(filePath):
    """
    Retrive all sheetnames of the given excel file path.
    

    Parameters
    ----------
    filePath : str
        Path of excel file to get sheet names.

    Returns
    -------
    sheetNameList : list of str
        

    """
    xlFile = pd.ExcelFile(filePath)
    sheetNameList = xlFile.sheet_names
    xlFile.close()
    return sheetNameList 


def getCountryExtract(countryList, sourceList='all'):
    """
    Extracts all available data in the database to an excel file
    """
    import datatoolbox as dt
    
    if sourceList == 'all':
        sourceList = list(dt.core.DB.sources.index)
    
    if not isinstance(countryList,list):
        countryList = [countryList]
    if not isinstance(sourceList,list):
        sourceList = [sourceList]
            
    #%%
    resFull = list()
    sourceList.sort()
    for source in tqdm.tqdm(sourceList):
        print(source)
        newList = list(dt.find(source=source).index)    
        newList.sort()
        resFull = resFull + newList

    _res2Excel(resFull, countryList)


def pandasStr2floatPercent(X):
    return([_str2float(x) for x in X])  
    
    
#%% Classes
    
class Excel_Reader():
    """
    Reader clase that uses an setup information provided in the excel file
    to extract data
    """
    
    def __init__(self, 
                 setup=None, 
                 fileName=None, 
                 overwrite=False, 
                 setupSheet = 'OUTPUT'):
        
        self.overwrite = overwrite
        self.setupSheet= setupSheet
        if setup:
            self.setup = setup
            fileSetup = True
        else:
            self.setup = dict()
            self.setup['fileName'] = fileName
            fileSetup = False
    
    def close(self):
        self.wb.close()
        
    def openSourceFile(self):
        if OS == 'Linux':
            os.system('libreoffice ' + self.setup['fileName'])
        elif OS == 'Darwin':
            os.system('open -a "Microsoft Excel" ' + self.setup['fileName'])        
   
    def getSetups(self):
        self.wb_read = load_workbook(self.setup['fileName'], data_only=True)
        self.wb = load_workbook(self.setup['fileName'], data_only=True)
        setup = dict()
        setup['fileName'] = self.setup['fileName']
        
        
        mapping = pd.read_excel(self.setup['fileName'], sheet_name =self.setupSheet)
        
        
        
        setupDict = {'sheetName'     : 'sheetName',
                     'timeIdxList'   : 'timeCells',
                     'spaceIdxList'  : 'spaceCells',
                     'dataID'        : 'dataCells',
                     'unit'          : 'unit',
                     'unitTo'        : 'unitTo'}

        setupDict = {'sheetName'     : 'Sheetname',
                     'timeIdxList'   : 'Time',
                     'spaceIdxList'  : 'Region',
                     'dataID'        : 'Variable',
                     'unit'          : 'Unit',
                     'unitTo'        : 'UnitTo',
                     'scenario'      : 'Scenario',
                     'source'        : 'Source'}
        
        for i,setupMapp in mapping.iterrows():
            for key in setupDict.keys():
                setup[key] = str(setupMapp[setupDict[key]])
                if setup[key] == 'nan':
                    setup[key] = None
                    
            yield(dict(setup))
            
#        setupSheet = self.wb[self.setupSheet]
#        for row in setupSheet.rows:
#            if row[0].value == 'sheet':
#                continue
#            
#            setup['sheetName']   = row[0].value
#            setup['timeIdxList']  = row[1].value
#            setup['spaceIdxList'] = row[2].value
#            setup['dataID']      = row[3].value
#            setup['unit']        = row[4].value
#            
#            for key in setup.keys():
#                if isinstance(setup[key],str):
#                    setup[key] = setup[key].replace('\ufeff','')
#            yield(setup)
        
    def read_data(self):
        replaceDict = {'EU': 'EU28'}
        from shutil import copyfile
        from copy import copy
        import mapping as mapp
        import datatoolbox as dt

        wb = load_workbook(self.setup['fileName'], data_only=True)
        
        validSpatialIDs = mapp.getValidSpatialIDs()
        self.setupList = list()
        
        tablesToReturn = dt.TableSet()
        
        for setup in self.getSetups():
            if isinstance(setup['timeIdxList'], int):
                setup['timeIdxList'] = str(setup['timeIdxList'])
            if isinstance(setup['spaceIdxList'], int):
                setup['spaceIdxList'] = str(setup['spaceIdxList'])
            print(setup)
            self.setupList.append(copy(setup))
            args =  None, None, None, None, None
            wksSheet = wb[setup['sheetName']]

            
            
            args =  [None, None, None, None, None]
            for argsTime in _iterTime(setup['timeIdxList'], wksSheet, *args):
                
                if argsTime[TI_ARG] is None:
                    print('No time defintion found')
                    print(setup['timeIdxList'])

                    continue
                print(argsTime[TI_ARG])
                print(setup['spaceIdxList'])
                for argsSpace in _iterSpace(setup['spaceIdxList'], wksSheet, *argsTime):
                    if argsSpace[SP_ARG] is None:
                        print('Not space defintion found')
                        continue
                    
                    if argsSpace[SP_ARG] not in validSpatialIDs:
                        print('No valid ISO code...')
                        
                        
                        if argsSpace[SP_ARG] in replaceDict.keys():
                            newID = replaceDict[argsSpace[SP_ARG]]
                        else:
                            newID = util.getCountryISO(argsSpace[SP_ARG])
                        

                        if newID in validSpatialIDs:
                            argsSpace[SP_ARG] = newID
                            print('Set iso code to: ' + argsSpace[SP_ARG])
                        else:
                            print('No ISO code found')
                            print(argsSpace[SP_ARG] + ' not found')
                            continue
                        
                    print(argsSpace)
                    for argData in _iterData(setup['dataID'], wksSheet, *argsSpace):
                        if argData[DT_ARG] is None:
                            print('No fitting dataID code found')
                            continue

                        meta = {'entity': argData[DT_ARG],
                                                      'unit' : setup['unit'],
                                                      'category':'',
                                                      'scenario' : setup['scenario'],
                                                      'source' : setup['source']}
#                        print(meta)
                        ID = core._createDatabaseID(meta)
                        
                        if ID not in tablesToReturn.keys():
                            table = Datatable()
                            table.meta = {'entity': argData[DT_ARG],
                                                      'unit' : setup['unit'],
                                                      'category':'',
                                                      'scenario' : setup['scenario'],
                                                      'source' : setup['source']}

                            tablesToReturn.add(table)
#                        print(argData)
                        value = self._readValue(wksSheet, xlsRow=argData[0], xlsCol=argData[1])

                        if isinstance(value, pd.DataFrame):
                            value = pandasStr2floatPercent(value)
                        else:
                            if value != '#VALUE!':
                                value = _str2float(value)
#                        print(value)
                        
#                        if  setup['unit'] == '%':
#                            value = value*100
#                        sdf
                        tablesToReturn[ID].loc[argData[SP_ARG],argData[TI_ARG]] = value
#                            print('success')
#                        iCount +=1
                            
#                        except Exception as e:
##                            print(e)
#                            pass
#                    self.wb.save(self.setup['fileName'])
#                            import pdb
#                            pdb.set_trace()
                        
        for table in tablesToReturn:
            table.columns = table.columns.astype(int)
            if '%' in table.meta['unit']:
                 table.loc[:,:] = table.loc[:,:]*100
                 tablesToReturn[table.ID] = table
                        
        return tablesToReturn
#            wb.save(self.setup['fileName'])
#            print('{} items inserted'.format(iCount))
#            
#            wb.close()

    def _readValue(self, wrkSheet, xlsRow, xlsCol):
#        if self.overwrite or pd.isna():
#            wrkSheet.cell(row=xlsRow, column=xlsCol, value = value)
        return wrkSheet.cell(row=xlsRow, column=xlsCol).value
#    def _writeMultipleIndicators(self, setup):
#        ws_readValues = load_workbook(self.setup['fileName'], data_only=True)[setup['sheetName']]
#        try:
#            tableIDs = [cell.value for cell in ws_readValues[setup['dataID']]]
#        except:
#            tableIDs = [cell[0].value for cell in ws_readValues[setup['dataID']]]
#            
#            tables = dt.getTables(tableIDs)
#            
#            REG_FIND_ROWS.search(setup['timeIdxList'])
        
class ExcelReader():
    
    def __init__(self, setupDict, xlsFile=None):
        """
        Required setup parameters:
            filePath    = './data/'
            fileName    = "test.xls"
            sheetName   = "Sheet0"
            timeIdxList  = ("B1", "C1")
            spaceIdxList = ('A2', 'A10')
        
        """
        self.setup(**setupDict)
        if xlsFile is not None:
            self.xlsFile = xlsFile
        
    
#    def setSetup(self, **kwargs):
#        for key in kwargs.keys():
#            self.setup[key] = kwargs[key]

    def setup(self, **kwargs):
        for key in kwargs.keys():
            
            
            if 'Idx' in key:
                # conversion of index
                setattr(self, key, [excel_to_pandas_idx(x) for x in kwargs[key]])
            else:
                setattr(self, key, kwargs[key])
        if hasattr(self,'df'):
            del self.df
#           
#        self.validate()
    
    def getAllSheetNames(self):
        xlFile = pd.ExcelFile(os.path.join(self.filePath, self.fileName))
        sheetNameList = xlFile.sheet_names
        xlFile.close()
        return sheetNameList

    def validate(self,):
        for attr in REQUIRED_SETUP_FIELDS:
            if not hasattr(self, attr):
                print(attr + " is missing")

    def gatherValue(self, excelIdx):
        if not hasattr(self, 'df'):
            if hasattr(self, 'xlsFile'):
                self.df = pd.read_excel(self.xlsFile, sheet_name=self.sheetName, header=None)
            else:
                self.df = pd.read_excel(os.path.join(self.filePath, self.fileName), sheet_name=self.sheetName, header=None)
        # elif core.config.DEBUG:
            # print('use loaded df')
        return self.df.iloc[excel_to_pandas_idx(excelIdx)]             

    def getAllData(self):
        return self.df
        
    def openFile(self):
        import os
        os.system('libreoffice ' + os.path.join(self.filePath, self.fileName)) 
                        
    def gatherData(self, load=True):

        if load:
            self.df = pd.read_excel(os.path.join(self.filePath, self.fileName), sheet_name=self.sheetName, header=None)
#            print('df loaded')
        elif core.config.DEBUG:
            print('use loaded df')

#        print(self.spaceIdxList)
#        print(self.timeIdxList)
        
        if len(self.timeIdxList) >1:
            timeIdx = self.df.iloc[self.timeIdxList[0][0],self.timeIdxList[0][1]:self.timeIdxList[1][1]+1]
            columns = self.df.columns[self.timeIdxList[0][1]:self.timeIdxList[1][1]+1]
#            print(timeIdx)
        else:
            timeIdx = self.df.iloc[self.timeIdxList[0][0],[self.timeIdxList[0][1]]]
            columns = self.df.columns[[self.timeIdxList[0][1]]]

        if len(self.spaceIdxList) >1:
            regions   = self.df.iloc[self.spaceIdxList[0][0]:self.spaceIdxList[1][0]+1,self.spaceIdxList[0][1]]
            index = self.df.index[self.spaceIdxList[0][0]:self.spaceIdxList[1][0]+1]
        else:
            regions   = self.df.iloc[[self.spaceIdxList[0][0]],self.spaceIdxList[0][1]]
            index = self.df.index[[self.spaceIdxList[0][0]]]
            
        if len(index) == 1:
#            print(index)
#            print(columns)
#            print(regions)
#            print(timeIdx)

            if len(columns) == 1:
                # extraction of a single value    
                data    = self.df.loc[index,columns]
                return pd.DataFrame(data.values, columns=regions, index=timeIdx).T
            
            else:
                # extraction of time serie
                data    = self.df.iloc[self.spaceIdxList[0][0],self.timeIdxList[0][1]:self.timeIdxList[1][1]+1]
                #print(type(data))
                #print(data.index)
                return pd.DataFrame(data.values, columns = regions, index = timeIdx).T
#                return pd.Series(data.values, index=columns) 
            
        else:
        
            if len(columns) == 1:
                # extraction of spatial series
                data    = self.df.loc[index,columns]
                
                return pd.DataFrame(data.values, columns=regions, index=timeIdx).T
#                return pd.Series(data.values, index=index) 
            else:
                # extraction of both spatial and time data
#                print(self.df)
#                print(index)
#                print(timeIdx)
#                print(columns)
                data    = self.df.loc[index,columns]
#                print(data)
                
                return pd.DataFrame(data.values, columns=timeIdx, index=regions)


  
class ExcelWriter():
    """
    More complex excel interface
    Author: Andreas Geiges
    """
    def __init__(self, setup=None, 
                 fileName=None, 
                 overwrite=False, 
                 setupSheet = 'INPUT',
                 interpolate=False):
        self.overwrite = overwrite # overwriting values
        self.setupSheet= setupSheet
        if setup:
            self.setup = setup
            fileSetup = True
        else:
            self.setup = dict()
            self.setup['fileName'] = fileName
            fileSetup = False
    
        self.interpolate = interpolate
    
    def close(self):
        self.wb.close()
        
        
    def getSetups(self):
        self.wb_read = load_workbook(self.setup['fileName'], data_only=True)
        self.wb = load_workbook(self.setup['fileName'])
        setup = dict()
        setup['fileName'] = self.setup['fileName']
        
        # new using pandas
        mapping = pd.read_excel(self.setup['fileName'], sheet_name=self.setupSheet)
        mapping.columns = [x.lower() for x in mapping.columns]
        
        
        setupDict = {'sheetName'     : 'sheetname',
                     'timeIdxList'   : 'time',
                     'spaceIdxList'  : 'region',
                     'dataID'        : 'variable',
                     'unit'          : 'unit',
#                     'unitTo'        : 'unitto'
                     }
        
        for i,setupMapp in mapping.iterrows():
            print(setupMapp)
            for key in setupDict.keys():
                setup[key] = str(setupMapp[setupDict[key]])
                if setup[key] == 'nan':
                    setup[key] = None
            yield(dict(setup))
        # old 
#        setupSheet = self.wb[self.setupSheet]
#        for row in setupSheet.rows:
#            if row[0].value == 'sheet':
#                continue
#            
#            setup['sheetName']   = row[0].value
#            setup['timeIdxList']  = row[1].value
#            setup['spaceIdxList'] = row[2].value
#            setup['dataID']      = row[3].value
#            setup['unit']        = row[4].value
#            
#            for key in setup.keys():
#                if isinstance(setup[key],str):
#                    setup[key] = setup[key].replace('\ufeff','')
#            yield(setup)
        
    def insert_data(self):
        replaceDict = {'EU': 'EU28'}
        from shutil import copyfile
        from copy import copy
        
        # Copy old file with "_filled" extension and load it
        if '.xlsx' in self.setup['fileName']:
            saveFileName = self.setup['fileName'].replace('.xlsx','_filled.xlsx')
        elif '.xls' in self.setup['fileName']:
            saveFileName = self.setup['fileName'].replace('.xls','_filled.xlsx')
        else:
            print('file extention not recognized')
        copyfile(self.setup['fileName'], saveFileName)
        self.setup['fileName'] = saveFileName
#        wb = load_workbook(self.setup['fileName'])
        
        
        # pre-load all valid spatial IDS
        validSpatialIDs = mapp.getValidSpatialIDs()
        
        # create list of setups
        self.setupList = list()
        
        #loop over all defined inputs in the INPUT sheet mapping (rows)
        for setup in self.getSetups():
            
            #ensure input is at all times a string
            if isinstance(setup['timeIdxList'], int):
                setup['timeIdxList'] = str(setup['timeIdxList'])
            if isinstance(setup['spaceIdxList'], int):
                setup['spaceIdxList'] = str(setup['spaceIdxList'])
            
            if config.DEBUG:
                print(setup)
            
            #add a copy of setup to list
            self.setupList.append(copy(setup))
            
            args =  None, None, None, None, None
            wksSheet_read  = self.wb_read[setup['sheetName']]
            wksSheet_write = self.wb[setup['sheetName']]
            # loop overall setups and collect dataIDs for pre-loading
            tableIDs = list()
            for args in _iterData(setup['dataID'], wksSheet_read, *args):
                
                if config.DEBUG:
                    print(args)
                if (~pd.isna(args[4])) and (args[4] is not None):
                    tableIDs.append(args[4])
            
            if config.DEBUG:
                print(tableIDs)
                
            #load all tables from database
            tables = core.DB.getTables(tableIDs)
            
            if setup['unit'] is not None:
                 for tableKey in tables.keys():
                     table = tables[tableKey]
                     tables[tableKey]= table.convert(setup['unit'])
                     
            if self.interpolate:
                 for tableKey in tables.keys():
                     for col in list(range(tables[tableKey].columns.min(),tables[tableKey].columns.max()+1)):
                         if col not in tables[tableKey].columns:
                             tables[tableKey].loc[:,col] = np.nan
                     tables[tableKey] = tables[tableKey].interpolate()    
                     
#            wksSheet = load_workbook(setup['fileName'], data_only=True)[setup['sheetName']]
            
            args =  [None, None, None, None, None]
            iCount = 0
            
            #iterate over time index list
            for argsTime in _iterTime(setup['timeIdxList'], wksSheet_read, *args):
                if argsTime[TI_ARG] is None:
                    print('No time defintion found')
                    continue
                
                #iterate over space index list
                for argsSpace in _iterSpace(setup['spaceIdxList'], wksSheet_read, *argsTime):
                    if argsSpace[SP_ARG] is None:
                        print('not spacial defintion found')
                        continue
                    
                    
                    if argsSpace[SP_ARG] not in validSpatialIDs:
#                        print('not time defintion found')
#                        print(argsSpace[SP_ARG])
                        if argsSpace[SP_ARG] in replaceDict.keys():
                            newID = replaceDict[argsSpace[SP_ARG]]
                        else:
                            newID = util.getCountryISO(argsSpace[SP_ARG])
                        

                        if newID in validSpatialIDs:
                            argsSpace[SP_ARG] = newID
                            print(argsSpace[SP_ARG])
                        else:
                            print(argsSpace[SP_ARG] + ' not found')
                            continue
                        
                    #print(argsSpace[SP_ARG])
                    #iterate over all data indices
                    for argData in _iterData(setup['dataID'], wksSheet_read, *argsSpace):
                        if argData[DT_ARG] is None:
                            print('not data defintion found')
                            continue

                        try:
                            value = tables[argData[DT_ARG]].loc[argData[SP_ARG],int(argData[TI_ARG])]
                            self._writeValueNew(wksSheet_write, xlsRow=argData[0], xlsCol=argData[1], value=value)
                            
#                            print('success')
                            iCount +=1
                            
                        except Exception as e:
                            if config.DEBUG:
                                print('failed with agruments:' + str(argData))
                                print(e)
                                #import pdb
                                #pdb.set_trace()
                            pass

            self.wb.save(self.setup['fileName'])
            print(setup)
            print('{} items inserted'.format(iCount))
            
            self.wb.close()

    def _writeValueNew(self, wrkSheet, xlsRow, xlsCol, value):
        if self.overwrite or pd.isna(wrkSheet.cell(row=xlsRow, column=xlsCol).value):
            print(value)
            wrkSheet.cell(row=xlsRow, column=xlsCol, value = value)
        
#    def _writeMultipleIndicators(self, setup):
#        ws_readValues = load_workbook(self.setup['fileName'], data_only=True)[setup['sheetName']]
#        try:
#            tableIDs = [cell.value for cell in ws_readValues[setup['dataID']]]
#        except:
#            tableIDs = [cell[0].value for cell in ws_readValues[setup['dataID']]]
#            
#            tables = dt.getTables(tableIDs)
#            
#            REG_FIND_ROWS.search(setup['timeIdxList'])
            
def createExcelInventory():
    """
    Create a excel inventory of the current data base. The excel files will be 
    placed in the directory of the database (see config.PATH_TO_DATASHELF)

    Returns
    -------
    None.

    """
    #%%
    import datatoolbox as dt
    import xlwings as xl
    import pandas as pd
    import time
    import os
    import tqdm
    import math
    xl.books.open(dt.config.PATH_TO_DATASHELF + '/excel_inventory/main.xlsx')
    try:
        wb = xl.books.open(dt.config.PATH_TO_DATASHELF + '/excel_inventory/main.xlsx')
    except:
        wb = xl.books.add()
    try:
        sht = wb.sheets.add('Sources')
    except:
        sht = wb.sheets['Sources']
    row = 6
    sheets = dict()
    sht.range('B2').value = 'Datatoolbox Inventory - ' + time.strftime("%Y/%m/%d")
    
    sht.range('B5').value = ['Source ID', 'Licence', 'Collected by', 'Url']
    sht.range('B5:E5').color = (0,104,204)
#    sht.range('B2').value
    for sourceID, source_data in tqdm.tqdm(dt.sourceInfo().loc['SDG_DB_2019':,:].iterrows()):
        
        sht.range('B'+str(row)).add_hyperlink('#{}!A1'.format(sourceID),sourceID)
        sht.range('C'+str(row)).value =  source_data.loc[ ['licence', 'collected_by', 'source_url']].values
        try:
            sheets[sourceID] = wb.sheets.add(sourceID, after=1)
        except:
            sheets[sourceID] = wb.sheets[sourceID]
        sheets[sourceID].range('A1').add_hyperlink('#Sources!A1','back to Index')
        sheets[sourceID].range('B2').value = sourceID + ' - Inventory'
        
        
        path = dt.config.PATH_TO_DATASHELF + 'excel_inventory/sources/'
        file_name = sourceID + '.xlsx'
        link = 'file://' + os.path.join(path + file_name) + '#Inventory!A1'
        
        sheets[sourceID].range('D2').add_hyperlink(link,'detailed inventory...')
        
        
        data_source = dt.find(source=sourceID)
#        pathways = list(data_source.pathway.unique())
        entities = sorted(list(data_source.variable.unique()))
        
        sheets[sourceID].range((5,2)).value = ['Variable', 'Entity', 'Category', 'Historic', 'Modeled']
        sheets[sourceID].range((5,2), (5,6)).color = (0,102,204)
        
        data_to_excel = pd.DataFrame(columns = ['Variable', 'Entity', 'Category', 'Historic', 'Modeled'])
        for index, entity in enumerate(entities):
#            
##            categories = sorted(list(dt.findp(source=sourceID).category.unique()))
            mask = data_source.variable == entity
            data = data_source.loc[data_source.index[mask],:]
            data_to_excel.loc[index,['Variable', 'Entity', 'Category']] = data.loc[data.index[0],['variable', 'entity', 'category']].values
            
            hist_scen = [x for x in data.pathway.unique() if ('Hist' in x) or ('hist' in x)]  
            data_to_excel.loc[index,['Historic']] = '{} hist pathways'.format(len(hist_scen)) 
            
            other_scen = [x for x in data.pathway.unique() if ('Hist' in x) or ('hist' in x)]  
            data_to_excel.loc[index,['Modeled']] = '{} other pathways'.format(len(data.pathway.unique()) - len(hist_scen)) 
            
        sheets[sourceID].range((6,2)).value = data_to_excel.values    
        sheets[sourceID].range((6,2), (30,6)).autofit()
        
        if row % 2 == 0:
            sht.range('B{0}:E{0}'.format(row)).color = (153,204,255)
        else:
            sht.range('B{0}:E{0}'.format(row)).color = (204,229,255)
        
        row +=1    
#        dsf
        _create_source_inventory(sourceID)

    sht.autofit()
#%%
def _create_source_inventory(sourceID):
    #%
    import os 
    import datatoolbox as dt
    import xlwings as xl
    import pandas as pd
    import time
    path = dt.config.PATH_TO_DATASHELF + '/excel_inventory/sources/'
    file_name = sourceID + '.xlsx'
    try:
        wb = xl.books.open(os.path.join(path, file_name))
    except:
        wb = xl.books.add()
        wb.save(os.path.join(path, file_name))
    try:
        sht = wb.sheets.add('Inventory')
    except:
        sht = wb.sheets['Inventory']
    sheets = dict()
    path_link = dt.config.PATH_TO_DATASHELF + 'excel_inventory/'
    file_name_link = 'main.xlsx'
    link = 'file://' + os.path.join(path_link + file_name_link) + '#{}!A1'.format(sourceID)
        
    sht.range('A1').add_hyperlink(link,'back to Main inventory')
    sht.range('B2').value = sourceID
    sht.range('D2').value = ['Created:', time.strftime("%Y/%m/%d")]
    sht.range('D3').value = ['Licence:', dt.sourceInfo().loc[sourceID,'licence']]
    
    sht.range('B5').value = 'Pathways:'
    sht.range('B5:G5').color = (0,104,208)
    data_source = dt.find(source=sourceID)
    row = 6
    for pathway in sorted(list(data_source.pathway.unique())):
        excel_pathway_string = pathway.replace('|','').replace('-','_').replace('.','')
        sht.range('B'+str(row)).add_hyperlink('#{}!A1'.format(excel_pathway_string[:30]),excel_pathway_string)
        if row % 2 == 0:
            sht.range('B{0}:G{0}'.format(row)).color = (153,204,255)
        else:
            sht.range('B{0}:G{0}'.format(row)).color = (204,229,255)
        
             
        try:
            sheets[excel_pathway_string] = wb.sheets.add(excel_pathway_string[:30], after=1)
        except:
            sheets[excel_pathway_string] = wb.sheets[excel_pathway_string[:30]]
            
        
        mask = data_source.pathway == pathway
        data = data_source.loc[data_source.index[mask],:]
        sheets[excel_pathway_string].range('B4').value = ['Entity', 'Category']
        sheets[excel_pathway_string].range('B5').value = data.sort_values(by='variable').loc[:,['entity', 'category']].values
        sheets[excel_pathway_string].range('D4').value = data.sort_values(by='variable').loc[:,['unit']]
        
        sheets[excel_pathway_string].autofit()
        sheets[excel_pathway_string].range('A1').add_hyperlink('#Inventory!A1','back to Index')
        sheets[excel_pathway_string].range('B2').value = [excel_pathway_string, ' - Inventory - ']
        sheets[excel_pathway_string].range('D4').value = 'Table ID'
        sheets[excel_pathway_string].range((4,2), (4,5)).color = (0,102,204)
        row +=1
    sht.autofit()
    wb.save(os.path.join(path, file_name))
    wb.close()
    #xl.Range('B1').add_hyperlink('file://Book2.xlsx#Sheet2!A1',"Sheet2")	

#import xlwings as xw
#from xlwings import Book, Range
#from datetime import datetime
#import numpy as np
#wb = Book('Book2')
#Range('A2').value = 12
#Range('A1').expand('table').value
#Range('B1').value = [[1],[2],[3],[4],[5]]
#Range('B').value
#
#xw.Range('B2').value = 222
#xw.Range('A2').expand('table').value     
#xw.Range('A2').expand('down').value    
#xw.Range('A2').options(ndim=2).value
#    
#
##%%
#app = xw.apps.active
#wb = app.books.active
#sht = wb.sheets.active
#
#used_range = sht.used_range.address
#
#Range(used_range).expand('table').value
#
#xw.Range('B1').add_hyperlink('[Book2]Sheet2!$A$1',"Sheet2")
#xw.Range('B1').add_hyperlink(':Sheet2.A1',"Sheet2")
#xw.Range('B1').add_hyperlink('file:///Users/andreasgeiges/Documents/Book2.xlsx#Sheet2!A1',"Sheet2")
# 
    
def read_active_excel_sheet():
    """
    Read the entire active excel sheet to a pandas dataframe.
    
    Return pandas.Dataframe
    """
    import xlwings as xw
    app = xw.apps.active
    book = app.books.active
    sheet = book.sheets[0]
    df = pd.DataFrame(sheet.used_range.value)
    return df


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 27 11:05:30 2020

@author: ageiges
"""
from .. import config
from . import  for_datatables 

from . import pandas
from . import excel

from . import matplotlib

#%% optional
if config.AVAILABLE_XARRAY:
    from . import xarray

if config.AVAILABLE_DOCX:
    from . import word

#from . import magicc6

from . import  pyam




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 16 13:16:08 2021

@author: ageiges
"""


mapping= dict()

#%% CD LINKS
mapping['CDLINKS_2021'] = dict()
mapping['CDLINKS_2021']['region']= {
             'World' : 'World',
             'OECD90 and EU (and EU candidate) countries' : 'R5_OECD',
             'Countries from the Reforming Ecomonies of the Former Soviet Union' : 'R5_REF',
             'Asian countries except Japan' : 'R5_ASIA',
             'Countries of the Middle East and Africa' : 'R5_MID',
             'Latin American countries' : 'R5_LAM',
             'Federative Republic of Brazil' : 'BRA',
             "People's Repulic of China" : 'CHN',
             'European Union (28 member countries)': 'EU28',
             'Republic of India' : 'IDN', 
             'State of Japan' : 'JPN',
             'Russian Federation' :'RUS', 
             'United States of America': 'USA'}


#%% ENGAGE
mapping['ENGAGE_2021'] = dict()
mapping['ENGAGE_2021']['region']= {
            'Asia (R5)':  'R5_ASIA',
            'China (R10)': 'R10_China',
            'European Union and Western Europe (R10)' : 'R10_EUROPE',
            'Latin America (R10)' : 'R10_LAM',
            'Latin America (R5)' : 'R5_LAM',
            'Middle East & Africa (R5)' : 'R5_MAF',
            'Middle East (R10)' : 'R10_MIDEAST',
            'North America (R10)' : 'R10_NAMERICA',
            'OECD & EU (R5)' : 'R10_EU_OECD',
            'Other (R10)' : 'R10_OTHER',
            'Other (R5)' :'R5_OTHER',
            'Other Asian countries (R10)' : 'R10_ASIA',
            'Pacific OECD (R10)' : 'R10_PACIFIC_OECD',
            'Reforming Economies (R10)': 'R10_REF',
            'Reforming Economies (R5)' : 'R5_REF',
            'South Asia (R10)' : 'R10_SOUTH_ASIA',
            'Sub-Saharan Africa (R10)' : 'R10_SUB_AFIRICA',
            'World' : 'World'}


#%% IPCC_AR5
mapping['IPCC_AR5'] = dict()
mapping['IPCC_AR5']['region'] ={
            'ASIA': 'R5_ASIA',
            'LAM': 'R5_LAM',
            'MAF': 'R5_MAF',
            'OECD90': 'R5_OECD90',
            'REF': 'R5_REF',
            'World': 'World'}

#%% IPCC_SR15
mapping['IPCC_SR15_2019'] = dict()
mapping['IPCC_SR15_2019']['region'] = {
            'World' : 'World',
            'R5ROWO' : 'R5_ROWO',
            'R5ASIA' : 'R5_ASIA',
            'R5LAM' : 'R5_LAM',
            'R5MAF' : 'R5_MAF',
            'R5OECD90+EU' : 'R5_OECD',
            'R5REF' : 'R5_REF'
            }




#%% NGFS 
mapping['NGFS_2021'] = dict()
mapping['NGFS_2021']['region']= {
             'World' : 'World',
             'OECD90 and EU (and EU candidate) countries' : 'R5_OECD',
             'Countries from the Reforming Ecomonies of the Former Soviet Union' : 'R5_REF',
             'Asian countries except Japan' : 'R5_ASIA',
             'Countries of the Middle East and Africa' : 'R5_MID',
             'Latin American countries' : 'R5_LAM'}




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Apr 22 14:49:27 2022

@author: ageiges
"""

import copy
import datatoolbox as dt

from util_for_testing import df, df2, sourceMeta
 
dt.admin.switch_database_to_testing()

def test_dataset_from_query():
    
    res = dt.findp(variable = ['Numbers|Ones',
                               'Numbers|Fives']) # find all
    ds = dt.DataSet.from_query(res)

def test_sel_methods():
    res = dt.findp(variable = ['Numbers|Ones',
                               'Numbers|Fives']) # find all
    ds = dt.DataSet.from_query(res)
    
    sub = ds.sel(scenario='Historic')
    assert dict(sub.dims) =={'year': 5, 'region': 4,'source': 1, 'model': 1}
    
    sub = ds.sel(region=0)
    assert dict(sub.dims) == {'year': 5, 'source': 1,'pathway': 1}
    
    sub = ds.sel(year=2012)
    assert dict(sub.dims) == {'region': 4, 'source': 1,'pathway': 1}

def test_unit_conversion():
    res = dt.findp(variable = ['Numbers|One',
                               'Numbers|Fives']) # find all
    ds = dt.DataSet.from_query(res)
    mm_data = ds['Numbers|Fives'].pint.to('mm')
    
    assert (mm_data.values == ds['Numbers|Fives'].values*1000).all()
    
    # test Emission unit
    test_array = ds['Numbers|Fives'].pint.dequantify().pint.quantify('Mt CO2 / yr')
    test_array = test_array.pint.to('kt CO2/ d')

    assert (test_array.values == 13.689253935660503).all()

if __name__ =='__main__':
    test_dataset_from_query()
    test_sel_methods()
    test_unit_conversion()


#!/usr/bin/env python
from __future__ import print_function

import os
import sys
import subprocess

from setuptools import setup, find_packages

#find_packages(exclude=['*personal.py', ])
def main():
    packages = [
        'datatoolbox',
#	'tools',
#	'data',
        ]
    pack_dir = {
        'datatoolbox': 'datatoolbox',
#        'tools':'datatoolbox/tools',
#	'data': 'datatoolbox/data'
    }
    package_data = {
        'datatoolbox': [
            'data/*',
            'data/SANDBOX_datashelf/*',
            'data/SANDBOX_datashelf/mappings/*',
            'tools/*',
            'tutorials/*.py',
            'pint_definitions.txt']}
    with open('README.md', encoding='utf-8') as f:
        long_description = f.read()
#    packages = find_packages()
    setup_kwargs = {
        "name": "datatoolbox",
        "description": 'The Python Data Toolbox',
        "long_description" : long_description,
        "long_description_content_type" :"text/markdown",
        "author": 'Andreas Geiges',
        "author_email": 'a.geiges@gmail.com',
        "url": 'https://gitlab.com/climateanalytics/datatoolbox',
        "packages": packages,
        "package_dir": pack_dir,
        "package_data" : package_data,
        "use_scm_version": {'write_to': 'datatoolbox/version.py'},
        "setup_requires": ['setuptools_scm'],
        "install_requires": [
            "appdirs",
            "pandas",
            "regex",
            "gitpython",
            "openscm_units",
            "pint",
            "pint-xarray",
            "pycountry",
            "fuzzywuzzy",
            "tqdm",
            "pyam-iamc",
            "networkx",
            "pint_xarray",
            "openpyxl",
            "xarray",
            "deprecated",],
        }
    rtn = setup(**setup_kwargs)

if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 14 16:58:04 2020

@author: ageiges
"""

import datatoolbox as dt

#%%
pattern = 'Electricity_generation|Solar**'
pattern = 'Electricity_generation|Hydro**'
#pattern = 'Electricity|generation|Wind**'
#pattern ='Electricity|generation|Renewables**'
#pattern = 'Primary_Energy|Total**'
#pattern = 'Primary_Energy|Gas**'
res = dt.findp(variable = pattern, source='IEA_WEB**')

tSet = dt.getTables(res.index)
xData = tSet.to_xarray(dimensions=['region', 'year', 'source'])

        #%%
import matplotlib.pyplot as plt
plt.clf()
coISO = 'CHN'
plt.plot(xData.year, xData.loc[coISO,:,:].values.squeeze())
plt.legend(xData.source.values)
plt.title(pattern + ' - ' + coISO)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 09:58:27 2019

@author: Andreas Geiges
"""
import numpy as np
from .data_structures import Datatable, TableSet
from . import config
import os
import pandas as pd
from . import core
from .util import isin

class _MAGGIC6():
    def read_MAGICC6_ScenFile(fileName, **kwargs):
        VALID_MASS_UNITS= {
                'Pt': 1e18,
                'Gt': 1e15,
                'Mt': 1e12,
                'kt': 1e9,
                't' : 1e6,
                'Pg': 1e15,
                'Tg': 1e12,
                'Gg': 1e9,
                'Mg': 1e6,
                'kg': 1e3,
                'g' : 1}
        fid = open(fileName,'r')
        nDataRows    = int(fid.readline().replace('/n',''))
        
        while True:
        #for i, line in enumerate(fid.readlines()):
            line = fid.readline()
            if line[:11] == '{0: >11}'.format('YEARS'):
                break
        #get first header line
        entities  = line.split()[1:]
        
        
        #reading units
        unitLine = fid.readline().split()[1:]
        #print(unitLine)
        
        # find correct component
        components = [GHG_data.findEntryIdx(entity) for entity in entities]
    
        units = [unit for unit in unitLine if unit[:2] in VALID_MASS_UNITS]        
        
        
        replacementDict = {'MtN2O-N' : 'Mt N'}
    
        units = [replacementDict.get(unit, unit) for unit in units]
    
        columns = [(x,y,z) for x,y,z in zip(entities, components, units)]
        entityFrame =  pd.DataFrame(columns=entities)
        entityFrame.columns = pd.MultiIndex.from_tuples(columns)
        
        entityFrame.columns.names=['NAME', 'COMP','UNIT']
    
        for i, line in enumerate(fid.readlines()):
            
            if i == nDataRows:
                break
            data = line.split()
            entityFrame.loc[int(data[0])] = np.asarray(data[1:])
    
        # TODO: CHange to a results list of datatables
        return entityFrame


    def read_MAGICC6_MATLAB_bulkout(pathName):
        
        temp_offset = 0.61
        df = pd.read_table(pathName, skiprows=23, header=0, delim_whitespace=True, index_col=0)
        df = df + temp_offset
        df.index = df.index.astype(int)
        return df
        #df.values[df.values<0] = np.nan
    #    
    #    meta = dict()
    #    meta['entity'] = 'global_temp'
    #    meta['unit']   = '°C'
    #    return Datatable(df, meta=meta)
    
    def read_MAGICC6_BINOUT(filePath):
        import pymagicc as pym
        reader = pym.io._BinaryOutReader(filePath)
    
        metaData, df = reader.read()
        
        data = df.pivot(index='region',columns='time', values='value')
        
        metaDict = dict()
        metaDict['entity'] = df.variable[0]
        metaDict['source'] = 'MAGICC6_calculation'
        metaDict['unit']   = None
    
        return Datatable(data, meta=metaDict)

class _primap():
    def read_PRIMAP_csv(fileName):
        
        metaMapping = {
                'entity': 'SHEET_ENTITY',
                'unit':'SHEET_UNIT',
                'category':'SHEET_NAME_CATEGORY',
                'scenario': 'SHEET_SCENARIO',
                'model' : 'SHEET_SOURCE'
                }
        
        allDf = pd.read_csv(fileName, usecols=[0,1], index_col=0)
        #print(os.path.basename(fileName))
    
        firstDataRow = allDf.loc['SHEET_FIRSTDATAROW','Unnamed: 1']
        
        #bugfix of old wrong formated PRIMAP files
        try:
            int(firstDataRow)
        except:
            firstDataRow = pd.np.where(allDf.index == "Countries\Years")[0][0]+3
       
        firstMetaRow = pd.np.where(allDf.index == "&SHEET_SPECIFICATIONS")[0][0]+1
        
        metaPrimap = dict()
        for row in range(firstMetaRow,firstDataRow):
            key = allDf.index[row]
            value = allDf.iloc[row,0]
            if key =='/':
                break
            metaPrimap[key] = value
        
        data = pd.read_csv(fileName, header=firstDataRow-2, index_col=0)
        
        meta = dict()
        for metaKey in metaMapping:
            
            if isinstance(metaMapping[metaKey], list):
                value = '_'.join(metaPrimap[x] for x in metaMapping[metaKey])
            else:
                value = metaPrimap[metaMapping[metaKey]]
            meta[metaKey] = value
        
        
        
        table = Datatable(data, meta=meta)
        table = table.loc[:, util.yearsColumnsOnly(table)]
        table.columns = table.columns.astype(int)
        return table#, data
    
    def read_PRIMAP_Excel(fileName, sheet_name= 0, xlsFile=None):
        if xlsFile is not None:
            xlsFile = xlsFile
        else:   
            xlsFile = pd.ExcelFile(fileName)
        allDf = pd.read_excel(xlsFile, sheet_name = sheet_name, usecols=[0,1], index_col=0)
        #print(os.path.basename(fileName))
    
        firstDataRow = allDf.loc['SHEET_FIRSTDATAROW','Unnamed: 1']
        
        #bugfix of old wrong formated PRIMAP files
        try:
            int(firstDataRow)
        except:
            firstDataRow = pd.np.where(allDf.index == "Countries\Years")[0][0]+3
            
        #print(firstDataRow)
        setup = dict()
        setup['filePath']  = os.path.dirname(fileName) +'/'
        setup['fileName']  = os.path.basename(fileName)
        setup['sheetName'] = sheet_name
        setup['timeIdxList']  = ('B' + str(firstDataRow-1), 'XX'+ str(firstDataRow-1))
        setup['spaceIdxList'] = ('A'+ str(firstDataRow), 'A1000')
        #print(setup)
        ex = ExcelReader(setup, xlsFile=xlsFile)
        data = ex.gatherData().astype(float)
        #return data
        meta = {'source': '',
                'entity': allDf.loc['SHEET_ENTITY','Unnamed: 1'],
                'unit': allDf.loc['SHEET_UNIT','Unnamed: 1'],
                'category': allDf.loc['SHEET_NAME_CATEGORY','Unnamed: 1'],
                'scenario': allDf.loc['SHEET_SCENARIO','Unnamed: 1'] + '|' + allDf.loc['SHEET_SOURCE','Unnamed: 1']}
        REG_ton = re.compile('^[GM]t')
        xx = REG_ton.search(meta['unit'])
        
        if xx:
            meta['unit'] = meta['unit'].replace(xx.group(0), xx.group(0) + ' ')
        
        table = Datatable(data, meta=meta)
        try:
            table.columns = table.columns.astype(int)
        except:
            print('warning: Columns could not be converted to int')
        return table

    def write_PRIMAP_Excel(data, fileName):
        
        if ~isinstance(data, TableSet):
            
            if isinstance(data, list):
                tableSet = TableSet(data)
                
            elif isinstance(data, (Datatable, pd.DataFrame)):
                tableSet = TableSet([data])
                
            else:
                raise(BaseException('Could not identifiy compatible data'))
                
            
            

class _matlab():
    def load_mat_file_as_dict(self, file_path):
        """
        Function to load a complex mat file as a dictionary
        Source for code: 
    
        Parameters
        ----------
        file_path : str
            Path to the .mat file to load
    
        Returns
        -------
        None.
    
        """
        from scipy.io import loadmat, matlab
        
        def _check_vars(d):
            """
            Checks if entries in dictionary are mat-objects. If yes
            todict is called to change them to nested dictionaries
            """
            for key in d:
                if isinstance(d[key], matlab.mio5_params.mat_struct):
                    d[key] = _todict(d[key])
                elif isinstance(d[key], np.ndarray):
                    d[key] = _toarray(d[key])
            return d
    
        def _todict(matobj):
            """
            A recursive function which constructs from matobjects nested dictionaries
            """
            d = {}
            for strg in matobj._fieldnames:
                elem = matobj.__dict__[strg]
                if isinstance(elem, matlab.mio5_params.mat_struct):
                    d[strg] = _todict(elem)
                elif isinstance(elem, np.ndarray):
                    d[strg] = _toarray(elem)
                else:
                    d[strg] = elem
            return d
    
        def _toarray(ndarray):
            """
            A recursive function which constructs ndarray from cellarrays
            (which are loaded as numpy ndarrays), recursing into the elements
            if they contain matobjects.
            """
            if ndarray.dtype != 'float64':
                elem_list = []
                for sub_elem in ndarray:
                    if isinstance(sub_elem, matlab.mio5_params.mat_struct):
                        elem_list.append(_todict(sub_elem))
                    elif isinstance(sub_elem, np.ndarray):
                        elem_list.append(_toarray(sub_elem))
                    else:
                        elem_list.append(sub_elem)
                return np.array(elem_list)
            else:
                return ndarray
    
        data = loadmat(file_path, struct_as_record=False, squeeze_me=True)
        return _check_vars(data)

class _EmissionModulePIK():
    
    
    N_ROWS_TO_ADD = 1
    ID_PARTS = ['SHEET_ENTITY', 
                'SHEET_CATEGORY', 
                'SHEET_CLASS', 
                'SHEET_TYPE', 
                'SHEET_SCENARIO', 
                'SHEET_SOURCE']
    metaDict = {'&SHEET_SPECIFICATIONS': '',
                'SHEET_CODE': '',
                'SHEET_CATEGORY': '',
                'SHEET_NAME_CATEGORY': '',
                'SHEET_ENTITY': '',
                'SHEET_TYPE': 'NET',
                'SHEET_CLASS': 'TOTAL',
                'SHEET_DESCR': '',
                'SHEET_NOTE': '',
                'SHEET_SOURCE': '',
                'SHEET_SCENARIO': '',
                'SHEET_FIRSTDATAROW': '',
                'SHEET_UNIT': '',
                'SHEET_DATATYPE': '',
                'SHEET_SUBSOURCE':''}
    
    categoryNameMapping = {'CATM1A' : 'Aviation',
                           'CATM1B' : 'Marine',
                           'CATM0EL': 'National_total_exlucding_LULUCF'}
            
    def write_tables(self, file, tables, sheet_names):
        #%%
        
        
        if isinstance(file, str):
            close_writer = True
            writer = pd.ExcelWriter(file, engine='xlsxwriter')
        else:
            close_writer = False
            # assume pandas excel writer
            writer = file
            
        if not isinstance(tables, list):
            tables = list(tables)
            sheet_names = list(sheet_names)
        
        
        def _validLettersAndNumbers(string):
            return string.isalpha() or string.isnumeric()
        def validStrings(input):
            return  ''.join(filter(_validLettersAndNumbers, input))
    #%%
        
        
        for table, sheet_name in zip(tables, sheet_names):
            header = pd.DataFrame(data=list(self.metaDict.values()), index= self.metaDict.keys(), columns=['values'])
            
            header.loc['SHEET_ENTITY'] = validStrings(table.meta['entity']).upper().replace('EMISSIONS','')
            header.loc['SHEET_CATEGORY'] = validStrings(table.meta['category']).upper()
            header.loc['SHEET_NAME_CATEGORY'] = self.categoryNameMapping[validStrings(table.meta['category']).upper()]
            header.loc['SHEET_SCENARIO'] = validStrings(table.meta['pathway']).upper()
            header.loc['SHEET_UNIT'] = table.meta['unit']
            header.loc['SHEET_SOURCE'] = validStrings(table.meta['source']).upper()
            firstDataRow = len(header) + self.N_ROWS_TO_ADD
            header.loc['SHEET_FIRSTDATAROW'] = firstDataRow +2
            header.loc['SHEET_CODE']   = '_'.join([header.loc[x,'values'] for x in self.ID_PARTS])
            
        
#            sheet_name = header.loc['SHEET_CODE'][0][:31]
            header.to_excel(writer, header=None, sheet_name= sheet_name)
            pd.DataFrame(table).to_excel(writer, startrow=firstDataRow, sheet_name= sheet_name)
        
        if close_writer:
            writer.close()
#        Excel.open('pandas_positioning.xlsx')

class Excel():
    
    def open(filePath):
        if config.OS == 'Linux':
            os.system('libreoffice ' + filePath )
        elif config.OS == 'Darwin':
            os.system('open -a "Microsoft Excel" ' + filePath )

# if config.AVAILABLE_XARRAY:
from datatoolbox.tools import xarray as _xr
class _Xarray():

    to_Xarray = _xr.to_XDataArray
        
 
#%%    
    
class IAMC_PYAM():
    
    def __init__(self):
        
        self.entityDict = self._mapped_entities()
        
        
        
    def _mapped_entities(self):
        """
        Entities that are commonly used. Keys are the original code and 
        values are the datatools equivalent
        """
        eDict = {
                # Emissions
                 'Emissions|CO2': 'Emissions|CO2',
                 'Emissions|CH4': 'Emissions|CH4',
                 'Emissions|NOx': 'Emissions|NOx',
                 'Emissions|NH3': 'Emissions|NH3',
                 'Emissions|CO': 'Emissions|CO',
                 'Emissions|BC': 'Emissions|BC',
                 'Emissions|N2O': 'Emissions|N2O',
                 'Emissions|KYOTOGHG_AR4': 'Emissions|KYOTOGHG_AR4',
                 'Emissions|KYOTOGHGAR4': 'Emissions|KYOTOGHG_AR4',
                 'Emissions|CO|AFOLU': 'Emissions|CO|AFOLU',
                 'Emissions|SF6': 'Emissions|SF6',
                 'Emissions|OC': 'Emissions|OC',
                 'Emissions|HFC': 'Emissions|HFC',
                 'Emissions|KYOTOGHG_AR5': 'Emissions|KYOTOGHG_AR5',
                 'Emissions|Sulfur': 'Emissions|Sulfur',
                 'Emissions|CO|Energy': 'Emissions|CO|Energy',
                 'Emissions|F-Gases': 'Emissions|F-Gases',
                 'Emissions|Sulfur|AFOLU': 'Emissions|Sulfur|AFOLU',
                 'Emissions|VOC': 'Emissions|VOC',
                 'Emissions|PFC': 'Emissions|PFC',
                 # Energy
                 'Primary_Energy': 'Primary_Energy',
                 'Secondary_Energy': 'Secondary_Energy',
                 'Final Energy' 
                 #Temperature assessment
                 'Global_mean_temperature': 'Global_mean_temperature',
                 #Misc
                 'GDP|MER': 'GDP|MER',
                 'GDP|PPP': 'GDP|PPP',
                 'Subsidies' : 'Subsidies',
                 'Investment': 'Investment',
                 'Population' : 'Population',
                 'Price|Carbon': 'Price|Carbon',
                 'Carbon_Sequestration': 'Carbon_Sequestration',
                 #Testing
                 'Numbers' : 'Numbers'
                }
        
        return eDict
    
    def _split_variable(self, variable):
        for key in self.entityDict.keys():
            if key in variable:
                category = variable.replace(key,'').lstrip('_').lstrip('|')
                entity = self.entityDict[key]
                return entity, category
    
    def from_IamDataFrame(self, idf):
        """Extracts a consistent time-series for a unique variable from ``idf``
    
        Parameters
        ----------
        idf : pyam.IamDataFrame
            pyam data frame from which to extract
            
        Returns
        -------
        TableSet 
        Returns the extracted data as tableset
        
        """
        
        timeseries = idf.timeseries()
        meta_columns = set(timeseries.index.names).difference({"region"})
        
        datatables = TableSet()
        for model, scenario in idf.index:
            pathway = f"{scenario}|{model}"
            extra_meta = idf.meta.loc[model, scenario].to_dict()

            for variable in idf.variable:
                data = timeseries.loc[isin(model=model, scenario=scenario, variable=variable)]
                if data.empty:
                    continue

                entity, category = self._split_variable(variable)
                meta = {
                    "pathway": pathway,
                    "entity": entity,
                    "category": category,
                    **extra_meta,
                    **{
                        k: v
                        for k, v in zip(data.index.names, data.index[0])
                        if k in meta_columns
                    }
                }
                table = Datatable(data.droplevel(list(meta_columns)), meta=meta)

                try:
                    tableKey = table.generateTableID()
                except:
                    tableKey =  '__'.join([variable, pathway])

                datatables[tableKey] = table.clean()
        
        return datatables
    
    def to_IamDataFrame(tableSet):
        
        import pyam
        long_table = tableSet.to_LongTable()
        long_table.index.name = None
        idf = pyam.IamDataFrame(pd.DataFrame(long_table))

        meta = pd.DataFrame([df.meta for df in self.values()])
        if 'model' not in meta:
            meta['model'] = ""
        if 'scenario' not in meta:
            meta['scenario'] = ""
        meta = (
            meta[
                pd.Index(['model', 'scenario', 'pathway'])
                .append(meta.columns[meta.columns.str.startswith('source')])
            ]
            .set_index(['model', 'scenario'])
            .drop_duplicates()
        )

        idf.meta = meta
        idf.reset_exclude()

        return idf
    
    
    def read_IAMC_table(self, iamcData):
        """
        Class to help convert iamcData input to homogeneous data tables
        
        Tables are split according the column variable of the iamcData table.
        The following colums are expected ['model', 'scenario', 'region', 'variable', 'unit']
        and a variable amount of year columns.
        The relationList maps which variable name is mapped to which output table.
        
        RelationDict Example:
        relationDict = {'Emissions|KYOTOGHGAR4|IPCM0EL' : {'entitiy' : 'Emissions|KYOTOGHGAR4',
                                                           'category': 'IPCM0EL'}}
        """
        
        print('this function is decribcated, please use from_idf')
        from types import SimpleNamespace
    #    import re
        
    #    YEAR_EXP = re.compile('^[0-9]{4}$')
    #    dataColumnsIds = [int(x) for x in iamcData.columns if YEAR_EXP.search(str(x)) is not None] 
        
        dataColumnsIds = iamcData.year
        
        outTables = list()
        
        for varName, mappDict in relationDict.items():
            
            if varName not in set(iamcData.variable):
                raise(BaseException('Required variable "{}" not found in input table'.format(varName)))
            ids = iamcData.data.loc[:,'variable'] == varName
            idx0 = iamcData.data.index[ids][0]
            mask
            dataExtract = iamcData.data.loc[ids, dataColumnsIds]
            dataExtract.meta = SimpleNamespace()
            dataExtract.index= iamcData.data.region[ids]
            
            # asserting that the unit, scenario and model data is only containing
            #  the same value
            assert iamcData.data.unit[ids].nunique() ==1
            assert iamcData.data.scenario[ids].nunique() ==1
            assert iamcData.data.model[ids].nunique() ==1    
            
            meta = dict()
            for key in mappDict.keys():
                meta[key]   = mappDict[key]
    #        meta['entity'] = varName
            meta['model']    = iamcData.loc[idx0,'model']
            meta['scenario'] = iamcData.loc[idx0,'scenario']
            meta['unit']     = iamcData.loc[idx0,'unit']
        
            outTables.append(Datatable(dataExtract, meta= meta))
        
        return outTables

def read_long_table(longDf, relationList):

    """
    Function to  convert long table input to homogeneous data tables
    
    Tables are split according the column variable of the iamcData table.
    The following colums are expected ['model', 'scenario', 'region', 'variable', 'unit']
    and a variable amount of year columns.
    The relationList maps which variable name is mapped to which output table.
    """
    import datatoolbox as dt
    import pandas as pd
    
    requiredColumns = set(['model', 'scenario','region','variable','unit', 'value', 'year'])

    outTables = list()
    
    if not set(longDf.columns) == requiredColumns:
        raise(BaseException('Input data must have this columns' + str(requiredColumns)))
    
    for varName in relationList:
        
        if varName not in list(longDf.loc[:,'variable']):
            raise(BaseException('Required variable "{}" not found in input table'.format(varName)))
        ids = longDf.loc[:,'variable'] == varName
        idx0 = longDf.index[ids][0]
        dataExtract = longDf.loc[ids, :].pivot(index='region', columns='year', values='value')
        #dataExtract.index= longDf.region[ids]
        dataExtract.columns = dataExtract.columns.astype(int)
        
        # asserting that the unit, scenario and model data is only containing
        #  the same value
        assert longDf.unit[ids].nunique() ==1
        assert longDf.scenario[ids].nunique() ==1
        assert longDf.model[ids].nunique() ==1
        
        meta = dict()
        meta['entity'] = varName
        meta['model']    = longDf.loc[idx0,'model']
        meta['scenario'] = longDf.loc[idx0,'scenario']
        meta['unit']     = longDf.loc[idx0,'unit']
    
        outTables.append(dt.Datatable(dataExtract, meta= meta))
    
    return outTables

matlab = _matlab()
emission_module = _EmissionModulePIK()
if config.AVAILABLE_XARRAY:
    xarray = _Xarray()

pyam = IAMC_PYAM()



if __name__ == '__main__':
    import pandas as pd
    longDf = pd.read_csv('data/long_test_data.csv')
    outTables = read_long_table(longDf, ['CH4|AGR', 'CH4|DOM'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 09:34:50 2020

@author: ageiges
"""
import numpy as np
# from ..util import identifyCountry
from .. import mapping as mapp

def addCountryNames(table, as_index=False):
    names = list()
    for idx in table.index:
        if idx in mapp.countries.codes.index:
            names.append(mapp.countries.codes.loc[idx,'name'])
        else:
            names.append(idx)
    if as_index:
        table.index = names    
    else:
        table.loc[:,'country_name'] = names
    return table

def convertIndexToISO(table, iso_type='alpha3'):
    from datatoolbox.util import identifyCountry
    """
    Convert index of a dataframe into iso codes.

    Parameters
    ----------
    table : pandas.Dataframe or dt.DataTable
        Index of thos table consists of country strings.
    iso : TYPE, optional
        Either 'alpha3', alpha2 or numISO. The default is 'alpha3'.

    Returns
    -------
    table :  pandas.Dataframe or dt.DataTable
        Return old dataframe with new iso index.

    """
    replaceDict = dict()
    
    for idx in table.index:
        iso = identifyCountry(idx)
        if iso is not None:
            replaceDict[idx] = iso
    table.index = table.index.map(replaceDict)           
    table = table.loc[~table.index.isna(),:]

    if iso_type == 'alpha2':
        table.index = mapp.countries.codes.loc[table.index,'alpha2']
    elif iso_type == 'numISO':
        table.index = mapp.countries.codes.loc[table.index,'numISO'].astype(int)
    return table

def add_standard(table, iso_type='alpha3'):
    """
    Convert index of a dataframe into iso codes.

    Parameters
    ----------
    table : pandas.Dataframe or dt.DataTable
        Index of thos table consists of country strings.
    iso : TYPE, optional
        Either 'alpha3', alpha2 or numISO. The default is 'alpha3'.

    Returns
    -------
    table :  pandas.Dataframe or dt.DataTable
        Return old dataframe with new iso index.

    """
    replaceDict = dict()
    
    for idx in table.index:
        iso = identifyCountry(idx)
        if iso is not None:
            replaceDict[idx] = iso
    table.index = table.index.map(replaceDict)           
    table = table.loc[~table.index.isna(),:]

    if iso_type == 'alpha2':
        table.index = mapp.countries.codes.loc[table.index,'alpha2']
    elif iso_type == 'numISO':
        table.index = mapp.countries.codes.loc[table.index,'numISO'].astype(int)
    return table


def yearsColumnsOnly(index):
    """
    Extracts from any given index only the index list that can resemble 
    as year 
    
    e.g. 2001
    """
    
    import re
    REG_YEAR   = re.compile('^[0-9]{4}$')
    
    newColumns = []
    for col in index:
        if REG_YEAR.search(str(col)) is not None:
            newColumns.append(col)
        else:
            try: 
                if ~np.isnan(col) and REG_YEAR.search(str(int(col))) is not None:
                #   test float string
                    newColumns.append(col)
            except:
                pass
    return newColumns

def index_availablity(dataframe,
                      index_list):
    
    available_idx = dataframe.index.intersection(index_list)
    missing_idx = set(index_list).difference(available_idx)
    
    return available_idx, missing_idx
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Andreas Geiges

TUTORIAL:       
      
"""

import datatoolbox as dt
import pandas as pd
import matplotlib.pyplot as plt

#%% seach for data
dt.find(entity = 'Emissions|CO2').source.unique()
source = 'WDI_2019'
res = dt.find(entity = 'Emissions|CO2',source=source)
res.scenario.unique()
res.entity.unique()
res.index

#%%load data
data = dict()
data['APAC']   = dt.getTable('Emissions|CO2||Energy|Total__Historic__APAC_2019')
data['UNFCCC'] = dt.getTable('Emissions|CO2||IPC0__Historic|CR__UNFCCC_CRF_2019')
data['WDI']    = dt.getTable('Emissions|CO2__Historic__WDI_2019')
data['PRIMAP'] = dt.getTable('Emissions|CO2|IPCM0EL__Historic|country_reported__PRIMAP_2019').convert('Mt CO2')
data['IEA']    = dt.getTable('Emissions|CO2|combustion|Total__Historic__IEA_CO2_FUEL_2019')


#%% plot data
coISO = 'USA'
plt.clf()
for key in data.keys():
    if coISO in data[key].index:
        plt.plot(data[key].columns,data[key].loc[coISO], label = key)
plt.legend()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug 18 13:57:45 2020

@author: ageiges
"""

import datatoolbox as dt
import numpy as np
import pandas as pd
import os

#def test_create_graph_basic():
#    from datatoolbox.utilities import get_data_trees,  plot_tree
#    kwargs = dict(model="MESSAGE", scenario="SSP2-19")
#    graphs, scenario = get_data_trees(**kwargs)
#    plot_tree(graphs["Emissions"], scenario, savefig_path=(os.path.join(os.getcwd(), "test.png")))
#
#def test_graph_from_search():
#    from datatoolbox.utilities import plot_query_as_graph
#    res = dt.find(entity = 'Area', source='WDI')
#    plot_query_as_graph(res)
#    
#
#def test_graph_integration():
#    dt.find(entity = 'Area', source='WDI').graph()

#if __name__ == '__main__':
#    test_create_graph_basic()
#    test_graph_from_search()
#    test_graph_integration()